from meltano.core.plugin import PluginInstall, PluginType
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.transform_add_service import TransformAddService
from meltano.core.behavior.hookable import hook


class DbtInvoker(PluginInvoker):
    def __init__(self, project, plugin, **kwargs):
        kwargs["run_dir"] = project.root_dir("transform")
        return super().__init__(project, plugin, **kwargs)

    def Popen_options(self):
        options = super().Popen_options()
        options["cwd"] = str(self.project.root_dir("transform"))

        return options


class DbtPlugin(PluginInstall):
    __plugin_type__ = PluginType.TRANSFORMERS
    __invoker_cls__ = DbtInvoker

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)


class DbtTransformPluginInstaller:
    def __init__(self, project, plugin):
        self.project = project
        self.plugin = plugin

    def install(self):
        # Add repo to my-test-project/transform/packages.yml
        transform_add_service = TransformAddService(self.project)
        transform_add_service.add_to_packages(self.plugin)

        # Add model and vars to my-test-project/transform/dbt_project.yml
        transform_add_service.update_dbt_project(self.plugin)


class DbtTransformPlugin(PluginInstall):
    __plugin_type__ = PluginType.TRANSFORMS
    __installer_cls__ = DbtTransformPluginInstaller

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)



import logging
import os
import re

from typing import Dict, List, Any
from sqlalchemy import create_engine
from snowflake.sqlalchemy import URL


# Don't show all the info log messages from Snowflake
for logger_name in ["snowflake.connector", "botocore", "boto3"]:
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.WARNING)


class SnowflakeConnector:
    def __init__(self, config: Dict = None) -> None:
        if not config:
            config = {
                "user": os.getenv("PERMISSION_BOT_USER"),
                "password": os.getenv("PERMISSION_BOT_PASSWORD"),
                "account": os.getenv("PERMISSION_BOT_ACCOUNT"),
                "database": os.getenv("PERMISSION_BOT_DATABASE"),
                "role": os.getenv("PERMISSION_BOT_ROLE"),
                "warehouse": os.getenv("PERMISSION_BOT_WAREHOUSE"),
            }

        self.engine = create_engine(
            URL(
                user=config["user"],
                password=config["password"],
                account=config["account"],
                database=config["database"],
                role=config["role"],
                warehouse=config["warehouse"],
                # Enable the insecure_mode if you get OCSP errors while testing
                # insecure_mode=True,
            )
        )

    def show_query(self, entity) -> List[str]:
        names = []

        query = f"SHOW {entity}"
        with self.engine.connect() as connection:
            results = connection.execute(query).fetchall()

            for result in results:
                names.append(result["name"].lower())

        return names

    def show_databases(self) -> List[str]:
        return self.show_query("DATABASES")

    def show_warehouses(self) -> List[str]:
        return self.show_query("WAREHOUSES")

    def show_roles(self) -> List[str]:
        return self.show_query("ROLES")

    def show_users(self) -> List[str]:
        return self.show_query("USERS")

    def show_schemas(self, database: str = None) -> List[str]:
        names = []

        if database:
            query = f"SHOW TERSE SCHEMAS IN DATABASE {database}"
        else:
            query = f"SHOW TERSE SCHEMAS IN ACCOUNT"

        with self.engine.connect() as connection:
            results = connection.execute(query).fetchall()

            for result in results:
                names.append(
                    f"{result['database_name'].lower()}.{result['name'].lower()}"
                )

        return names

    def show_tables(self, database: str = None, schema: str = None) -> List[str]:
        names = []

        if schema:
            query = f"SHOW TERSE TABLES IN SCHEMA {schema}"
        elif database:
            query = f"SHOW TERSE TABLES IN DATABASE {database}"
        else:
            query = f"SHOW TERSE TABLES IN ACCOUNT"

        with self.engine.connect() as connection:
            results = connection.execute(query).fetchall()

            for result in results:
                names.append(
                    f"{result['database_name'].lower()}"
                    + f".{result['schema_name'].lower()}"
                    + f".{result['name'].lower()}"
                )

        return names

    def show_views(self, database: str = None, schema: str = None) -> List[str]:
        names = []

        if schema:
            query = f"SHOW TERSE VIEWS IN SCHEMA {schema}"
        elif database:
            query = f"SHOW TERSE VIEWS IN DATABASE {database}"
        else:
            query = f"SHOW TERSE VIEWS IN ACCOUNT"

        with self.engine.connect() as connection:
            results = connection.execute(query).fetchall()

            for result in results:
                names.append(
                    f"{result['database_name'].lower()}"
                    + f".{result['schema_name'].lower()}"
                    + f".{result['name'].lower()}"
                )

        return names

    def show_future_grants(self, database: str = None, schema: str = None) -> List[str]:
        future_grants = {}

        if schema:
            query = f"SHOW FUTURE GRANTS IN SCHEMA {schema}"
        elif database:
            query = f"SHOW FUTURE GRANTS IN DATABASE {database}"
        else:
            pass

        with self.engine.connect() as connection:
            results = connection.execute(query).fetchall()

            for result in results:
                if result["grant_to"] == "ROLE":
                    role = result["grantee_name"].lower()
                    privilege = result["privilege"].lower()
                    granted_on = result["grant_on"].lower()

                    future_grants.setdefault(role, {}).setdefault(
                        privilege, {}
                    ).setdefault(granted_on, []).append(result["name"].lower())

                else:
                    continue

        return future_grants

    def show_grants_to_role(self, role) -> Dict[str, Any]:
        grants = {}

        query = f"SHOW GRANTS TO ROLE {SnowflakeConnector.snowflaky(role)}"
        with self.engine.connect() as connection:
            results = connection.execute(query).fetchall()

            for result in results:
                privilege = result["privilege"].lower()
                granted_on = result["granted_on"].lower()

                grants.setdefault(privilege, {}).setdefault(granted_on, []).append(
                    result["name"].lower()
                )

        return grants

    def show_roles_granted_to_user(self, user) -> List[str]:
        roles = []

        query = f"SHOW GRANTS TO USER {SnowflakeConnector.snowflaky(user)}"
        with self.engine.connect() as connection:
            results = connection.execute(query).fetchall()

            for result in results:
                roles.append(result["role"].lower())

        return roles

    def run_query(self, query: str):

        with self.engine.connect() as connection:
            result = connection.execute(query)

        return result

    def full_schema_list(self, schema: str) -> List[str]:
        """
        For a given schema name, get all schemas it may be referencing.

        For example, if <db>.* is given then all schemas in the database 
        will be returned. If <db>.<schema_partial>_* is given, then all 
        schemas that match the schema partial pattern will be returned. 
        If a full schema name is given, it will return that single schema
        as a list.

        This function can be enhanced in the future to handle more 
        complicated schema names if necessary.

        Returns a list of schema names.
        """
        # Generate the information_schema identifier for that database
        # in order to be able to filter it out
        name_parts = schema.split(".")

        info_schema = f"{name_parts[0]}.information_schema"

        fetched_schemas = []

        # All Schemas
        if name_parts[1] == "*":
            db_schemas = self.show_schemas(name_parts[0])
            for db_schema in db_schemas:
                if db_schema != info_schema:
                    fetched_schemas.append(db_schema)

        # Prefix schema match
        elif "*" in name_parts[1]:
            db_schemas = self.show_schemas(name_parts[0])
            for db_schema in db_schemas:
                schema_name = db_schema.split(".", 1)[1].lower()
                if schema_name.startswith(name_parts[1].split("*", 1)[0]):
                    fetched_schemas.append(db_schema)

        # TODO Handle more complicated matches

        else:
            # If no * in name, then return provided schema name
            fetched_schemas = [schema]

        return fetched_schemas

    def snowflaky(name: str) -> str:
        """
        Convert an entity name to an object identifier that will most probably be
        the proper name for Snowflake.

        e.g. gitlab-ci --> "gitlab-ci"
             527-INVESTIGATE$ISSUES.ANALYTICS.COUNTRY_CODES -->
             --> "527-INVESTIGATE$ISSUES".ANALYTICS.COUNTRY_CODES;

        Pronounced /snəʊfleɪkɪ/ like saying very fast snowflak[e and clarif]y
        Permission granted to use snowflaky as a verb.
        """
        name_parts = name.split(".")
        new_name_parts = []

        for part in name_parts:
            if (
                re.match("^[0-9a-zA-Z_]*$", part) is None  # Proper formatting
                and re.match('^".*"$', part) is None  # Already quoted
            ):
                new_name_parts.append(f'"{part}"')
            else:
                new_name_parts.append(part)

        return ".".join(new_name_parts)



from pathlib import Path


class Package:
    def __init__(self, name, project):
        self.name = name
        self.project = project

    @property
    def tables(self):
        return self.package_model_dir(".table")

    @property
    def topics(self):
        return self.package_model_dir(".topic")

    @property
    def files(self):
        return self.project.model_dir().glob(f"{self.name}/**/*.m5o")

    def package_model_dir(self, file_type):
        return [
            Path(f)
            for f in self.project.model_dir().glob(f"{self.name}/**/*{file_type}.m5o")
        ]



import os
import logging
import datetime

from meltano.core.utils import truthy


# Flask
# -----------------
THREADS_PER_PAGE = 1
PROFILE = truthy(os.getenv("FLASK_PROFILE"))

## Change this value in production
SECRET_KEY = "thisisnotapropersecretkey"

# Meltano
# -----------------
AIRFLOW_DISABLED = truthy(os.getenv("MELTANO_DISABLE_AIRFLOW"))
MELTANO_AUTHENTICATION = truthy(os.getenv("MELTANO_AUTHENTICATION"))
MELTANO_READONLY = truthy(os.getenv("MELTANO_READONLY"))
MELTANO_UI_URL = os.getenv("MELTANO_UI_URL", "/")
MELTANO_OAUTH_SERVICE_URL = os.getenv("MELTANO_OAUTH_SERVICE_URL", None)

API_ROOT_DIR = os.path.abspath(os.path.dirname(__file__))
TEMP_FOLDER = os.path.join(API_ROOT_DIR, "static/tmp")
PROJECT_ROOT_DIR = os.path.dirname(API_ROOT_DIR)

JSON_SCHEME_HEADER = "X-JSON-SCHEME"
VERSION_HEADER = "X-MELTANO-VERSION"

# Flask-SQLAlchemy
# -----------------
SQLALCHEMY_ECHO = False
SQLALCHEMY_TRACK_MODIFICATIONS = False
SQLALCHEMY_DATABASE_URI = os.getenv("MELTANO_DATABASE_URI")

# Flask-security
# -----------------

# Change this value in production
# A better approach would be to have individual salts hashed resource
SECURITY_PASSWORD_SALT = "b4c124932584ad6e69f2774a0ae5c138"
SECURITY_PASSWORD_HASH = "bcrypt"
SECURITY_REGISTERABLE = False
SECURITY_CHANGEABLE = True
SECURITY_RECOVERABLE = False
SECURITY_CONFIRMABLE = False
SECURITY_URL_PREFIX = "/auth"
SECURITY_USER_IDENTITY_ATTRIBUTES = ("username", "email")
SECURITY_SEND_REGISTER_EMAIL = False
SECURITY_SEND_PASSWORD_CHANGE_EMAIL = False

SECURITY_MSG_USERNAME_NOT_PROVIDED = ("You must provide a username.", "error")
SECURITY_MSG_USERNAME_INVALID = (
    "Username must be at least 6 characters long.",
    "error",
)
SECURITY_MSG_USERNAME_ALREADY_TAKEN = ("This username is already taken.", "error")

# Flask-Mail
# -----------------

# Change these configuration for your SMTP server
MAIL_SERVER = "localhost"
MAIL_PORT = 587
MAIL_DEFAULT_SENDER = '"Meltano" <bot@metlano.com>'

# Flask-Authlib
# -----------------

GITLAB_CLIENT_ID = os.getenv("OAUTH_GITLAB_APPLICATION_ID")
GITLAB_CLIENT_SECRET = os.getenv("OAUTH_GITLAB_SECRET")

# Flask-RESTful
# -----------------

RESTFUL_JSON = {}

# Flask-Executor
# -----------------

EXECUTOR_PROPAGATE_EXCEPTIONS = True


# Flask-CORS
# -----------------

CORS_EXPOSE_HEADERS = [VERSION_HEADER]
CORS_ALLOW_HEADERS = ["CONTENT-TYPE", JSON_SCHEME_HEADER]


class Production(object):
    SESSION_COOKIE_SECURE = True
    SESSION_COOKIE_HTTPONLY = True


def ensure_secure_setup(app):
    secure_variables = ["SERVER_NAME", "SECRET_KEY", "SECURITY_PASSWORD_SALT"]

    facts = []
    for var in secure_variables:
        if app.config[var] is None:
            facts.append(f"\t- '{var}': variable is unset.")
        elif app.config[var] == globals().get(var):
            facts.append(f"\t- '{var}': variable has test value.")

    if facts:
        facts_msg = "\n".join(facts)
        logging.warning(
            "The following variables are insecure and should be regenerated:\n"
            f"{facts_msg}\n\n"
            "Use `meltano ui setup` command to generate new secrets."
        )



import logging
import click
import sys

from meltano.core.permissions import grant_permissions, SpecLoadingError
from meltano.core.tracking import GoogleAnalyticsTracker
from . import cli
from .params import project


@cli.group()
def permissions():
    """Database permission related commands."""
    pass


@permissions.command()
@click.argument("spec")
@click.option(
    "--db",
    help="The type of the target DB the specifications file is for.",
    type=click.Choice(["postgres", "snowflake"]),
    required=True,
)
@click.option("--dry", help="Do not actually run, just check.", is_flag=True)
@click.option(
    "--diff", help="Show full diff, both new and existing permissions.", is_flag=True
)
@project()
def grant(project, db, spec, dry, diff):
    """Grant the permissions provided in the provided specification file."""
    try:
        sql_commands = grant_permissions(db, spec, dry_run=dry)
        tracker = GoogleAnalyticsTracker(project)
        tracker.track_meltano_permissions_grant(db=db, dry=dry)

        click.secho()
        if diff:
            click.secho(
                "SQL Commands generated for given spec file (Full diff with both new and already granted commands):"
            )
        else:
            click.secho("SQL Commands generated for given spec file:")
        click.secho()

        diff_prefix = ""
        for command in sql_commands:
            if command["already_granted"]:
                if diff:
                    diff_prefix = "  "
                else:
                    continue
            else:
                if diff:
                    diff_prefix = "+ "

            if command.get("run_status"):
                fg = "green"
                run_prefix = "[SUCCESS] "
            elif command.get("run_status") is None:
                fg = "cyan"
                run_prefix = "[SKIPPED] "
            else:
                fg = "red"
                run_prefix = "[ERROR] "

            click.secho(f"{diff_prefix}{run_prefix}{command['sql']};", fg=fg)
    except SpecLoadingError as exc:
        for line in str(exc).splitlines():
            click.secho(line, fg="red")
        sys.exit(1)



import subprocess
import platform
import sys
import os
from pathlib import Path
from collections import namedtuple

from .project import Project
from .error import SubprocessError


VenvSpecs = namedtuple("VenvSpecs", ("lib_dir", "bin_dir", "site_packages_dir"))


POSIX = VenvSpecs(
    lib_dir="lib",
    bin_dir="bin",
    site_packages_dir=os.path.join("lib", f"python{sys.version[:3]}", "site-packages"),
)

NT = VenvSpecs(
    lib_dir="Lib",
    bin_dir="Scripts",
    site_packages_dir=os.path.join("Lib", "site-packages"),
)


class VirtualEnv:
    PLATFORM_SPECS = {"Linux": POSIX, "Darwin": POSIX, "Windows": NT}

    def __init__(self, root: Path):
        self.root = root
        self._specs = self.__class__.platform_specs()

    @classmethod
    def platform_specs(cls):
        try:
            system = platform.system()
            return cls.PLATFORM_SPECS[system]
        except KeyError:
            raise Exception(f"Platform {system} not supported.")

    def __getattr__(self, attr):
        spec = getattr(self._specs, attr)
        return self.root.joinpath(spec)

    def __str__(self):
        return str(self.root)


class VenvService:
    def __init__(self, project):
        self.project = project

    def create(self, namespace="", name=""):
        venv = VirtualEnv(self.project.venvs_dir(namespace, name))

        run = subprocess.run(
            [sys.executable, "-m", "venv", str(venv)],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        if run.returncode != 0:
            raise SubprocessError(
                f"Could not create of the virtualenv for '{namespace}/{name}'", run
            )

        # we want the plugin to inherit our current venv
        sys_paths = []
        for path in sys.path:
            # the current venv if any
            if path.endswith("site-packages"):
                sys_paths.append(path)

            # when meltano is installed as editable
            if path.endswith(os.path.join("meltano", "src")):
                sys_paths.append(path)

        # inject a .pth to add these PYTHONPATHs
        meltano_pth_path = venv.site_packages_dir.joinpath("meltano_venv.pth")
        with meltano_pth_path.open("w") as pth:
            for path in sys_paths:
                pth.write(path + "\n")

        return run

    def install(self, pip_url, namespace="", name=""):
        """
        Install a package using `pip` in the proper virtual environment.

        This method is not threadsafe.
        """

        venv = VirtualEnv(self.project.venvs_dir(namespace, name))
        pip_install_path = venv.bin_dir.joinpath("pip")
        meltano_pth_path = venv.site_packages_dir.joinpath("meltano_venv.pth")
        meltano_pth_disable_path = meltano_pth_path.with_suffix(".pth.disable")

        try:
            # disable the `meltano venv`
            if meltano_pth_path.exists():
                meltano_pth_path.rename(meltano_pth_disable_path)

            run = subprocess.run(
                [str(pip_install_path), "install", *pip_url.split(" ")],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )

            if run.returncode != 0:
                raise SubprocessError(f"Failed to install plugin '{name}'.", run)
        finally:
            # re-enable the `meltano venv`
            if meltano_pth_disable_path.exists():
                meltano_pth_disable_path.rename(meltano_pth_path)

        return run

    def exec_path(self, bin, name=None, namespace=""):
        name = name or bin
        venv = VirtualEnv(self.project.venvs_dir(namespace, name))
        return venv.bin_dir.joinpath(bin)



import os
import yaml
import click
import shutil
import logging
from functools import singledispatch
from typing import List, Dict
from pathlib import Path

import meltano.core.bundle as bundle
from .project import Project
from .venv_service import VenvService


class ProjectInitServiceError(Exception):
    pass


@singledispatch
def visit(node, executor):
    pass


@visit.register(dict)
def _(node: Dict, target_path: Path = None):
    created = []

    logging.debug(f"{target_path}")
    for name, definition in node.items():
        directory = target_path.joinpath(os.path.dirname(name))

        # always create the base directory
        os.makedirs(directory, exist_ok=True)

        # recurse for the nested definition
        created += visit(definition, target_path.joinpath(name))

    return created


@visit.register(str)
def _(node: str, target_path: Path):
    """
    Create the file using either the raw content or a bundled file.
    """
    logging.debug(f"{target_path}")
    if node.startswith("bundle://"):
        # copy from the bundle
        _, path = node.split("bundle://")
        path = bundle.find(path)

        logging.debug(f"{path} → {target_path}")
        if path.is_dir():
            shutil.copytree(path, target_path)
        else:
            shutil.copy(path, target_path)
    else:
        # write the content
        with target_path.open("w") as target:
            target.write(node)

    return [target_path]


class ProjectInitService:
    def __init__(self, project_name):
        self.initialize_file = bundle.find("initialize.yml")
        self.project_name = project_name.lower()

    def init(self) -> Project:
        default_project_yaml = yaml.safe_load(open(self.initialize_file))
        try:
            os.mkdir(self.project_name)
        except Exception as e:
            raise ProjectInitServiceError

        new_project = Project(self.project_name)
        for path in visit(default_project_yaml, Path(self.project_name)):
            click.secho(f"\tCreated", fg="blue", nl=False)
            click.echo(f" {path}")

        return new_project

    def echo_instructions(self):
        click.secho("Project", nl=False)
        click.secho(f" {self.project_name}", fg="green", nl=False)
        click.echo(" has been created.")

        click.echo("\nNext steps:")
        click.secho("\tcd ", nl=False)
        click.secho(self.project_name, fg="green")
        click.echo(
            "\tVisit https://meltano.com/docs/getting-started.html in order to try us out"
        )

        click.secho("> ", fg="bright_black", nl=False)

        click.secho("Meltano sends anonymous usage data that help improve the product.")

        click.secho("> ", fg="bright_black", nl=False)

        click.echo("You can opt-out for new, existing, or all projects.")

        click.secho("> ", fg="bright_black", nl=False)

        click.secho(
            "https://meltano.com/docs/environment-variables.html#anonymous-usage-data",
            fg="cyan",
        )

    def join_with_project_base(self, filename):
        return os.path.join(".", self.project_name, filename)



from flask import jsonify, request

from .embeds_helper import EmbedsHelper, InvalidEmbedToken
from meltano.api.api_blueprint import APIBlueprint
from meltano.api.models import db


embedsBP = APIBlueprint("embeds", __name__)


@embedsBP.errorhandler(InvalidEmbedToken)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"No matching resource found or this resource is no longer public.",
            }
        ),
        400,
    )


@embedsBP.route("/embed/<token>", methods=["GET"])
def get_embed(token):
    embeds_helper = EmbedsHelper()
    response_data = embeds_helper.get_embed_from_token(db.session, token)

    return jsonify(response_data)


@embedsBP.route("/embed", methods=["POST"])
def embed():
    post_data = request.get_json()
    resource_id = post_data["resource_id"]
    resource_type = post_data["resource_type"]
    embeds_helper = EmbedsHelper()
    response_data = embeds_helper.generate_embed_snippet(
        db.session, resource_id, resource_type
    )

    return jsonify(response_data)



import sqlalchemy
from flask import jsonify, request

from .errors import InvalidFileNameError

from meltano.core.project import Project
from meltano.core.m5o.reports_service import ReportAlreadyExistsError, ReportsService

from meltano.api.api_blueprint import APIBlueprint
from meltano.api.security.resource_filter import ResourceFilter, NameFilterMixin, Need
from meltano.api.security.readonly_killswitch import readonly_killswitch

reportsBP = APIBlueprint("reports", __name__)


def reports_service():
    project = Project.find()
    return ReportsService(project)


class ReportFilter(NameFilterMixin, ResourceFilter):
    def __init__(self, *args):
        super().__init__(*args)

        self.needs(self.design_need)

    def design_need(self, permission_type, report):
        if permission_type == "view:reports":
            return Need("view:design", report["design"])


@reportsBP.errorhandler(ReportAlreadyExistsError)
def _handle(ex):
    report_name = ex.report["name"]
    return (
        jsonify(
            {
                "error": True,
                "code": f"A report with the name '{report_name}' already exists. Try renaming the report.",
            }
        ),
        409,
    )


@reportsBP.errorhandler(InvalidFileNameError)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"The report name provided is invalid. Try a name without special characters.",
            }
        ),
        400,
    )


@reportsBP.route("/", methods=["GET"])
def index():
    reports = reports_service().get_reports()
    reports = ReportFilter().filter_all("view:reports", reports)
    return jsonify(reports)


@reportsBP.route("/save", methods=["POST"])
@readonly_killswitch
def save_report():
    post_data = request.get_json()
    response_data = reports_service().save_report(post_data)
    return jsonify(response_data)


@reportsBP.route("/update", methods=["POST"])
@readonly_killswitch
def update_report():
    post_data = request.get_json()
    response_data = reports_service().update_report(post_data)
    return jsonify(response_data)



from typing import List

from meltano.core.permissions.pg_spec_loader import PGSpecLoader
from meltano.core.permissions.snowflake_spec_loader import SnowflakeSpecLoader
from meltano.core.permissions.utils.error import SpecLoadingError
from meltano.core.permissions.utils.snowflake_connector import SnowflakeConnector


def grant_permissions(db: str, spec_path: str, dry_run: bool) -> List[str]:
    if db == "postgres":
        spec_loader = PGSpecLoader(spec_path)
    elif db == "snowflake":
        spec_loader = SnowflakeSpecLoader(spec_path)
    else:
        raise SpecLoadingError(f"Permissions Spec File for {db} is not supported.")

    sql_grant_queries = spec_loader.generate_permission_queries()

    run_queries = []

    if not dry_run:
        conn = SnowflakeConnector()
        for query in sql_grant_queries:
            status = None
            if not query.get("already_granted"):
                try:
                    result = conn.run_query(query.get("sql"))
                    outcome = result.fetchall()
                    status = True
                except:
                    status = False

                ran_query = query
                ran_query["run_status"] = status
                run_queries.append(ran_query)
    else:
        run_queries = sql_grant_queries

    # Always return the SQL commands that ran
    return run_queries



import functools
from flask import current_app

from meltano.core.utils import truthy


HTTP_KILLSWITCH_CODE = 499


def readonly_killswitch(f):
    @functools.wraps(f)
    def _decorated(*args, **kwargs):
        if current_app.config["MELTANO_READONLY"]:
            return (
                "Meltano is currently running in read-only mode.",
                HTTP_KILLSWITCH_CODE,
            )

        return f(*args, **kwargs)

    return _decorated



import configparser
import shutil
import logging
import subprocess
import time
import os
from . import PluginInstall, PluginType

from meltano.core.behavior.hookable import hook
from meltano.core.plugin.config_service import PluginConfigService
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_invoker import invoker_factory, PluginInvoker
from meltano.core.db import project_engine
from meltano.core.venv_service import VenvService
from meltano.core.utils import nest, map_dict


class AirflowInvoker(PluginInvoker):
    def Popen_options(self):
        env = os.environ.copy()
        venv_dir = self.project.venvs_dir(self.plugin.type, self.plugin.name)

        # add the Airflow virtualenv because it contains `gunicorn`
        env["PATH"] = os.pathsep.join([str(venv_dir.joinpath("bin")), env["PATH"]])
        env["VIRTUAL_ENV"] = str(venv_dir)
        env["AIRFLOW_HOME"] = str(self.config_service.run_dir)

        options = super().Popen_options()
        options_env = nest(options, "env")
        options_env.update(env)

        return options


class Airflow(PluginInstall):
    __plugin_type__ = PluginType.ORCHESTRATORS
    __invoker_cls__ = AirflowInvoker

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)

    @property
    def config_files(self):
        return {"config": "airflow.cfg"}

    @hook("before_install")
    def setup_env(self, project, args=[]):
        # to make airflow installables without GPL dependency
        os.environ["SLUGIFY_USES_TEXT_UNIDECODE"] = "yes"

    @hook("after_install")
    def after_install(self, project, args=[]):
        _, Session = project_engine(project)
        session = Session()

        plugin_config_service = PluginConfigService(
            self,
            config_dir=project.plugin_dir(self),
            run_dir=project.run_dir(self.name),
        )

        plugin_settings_service = PluginSettingsService(project)
        airflow_cfg_path = plugin_config_service.run_dir.joinpath("airflow.cfg")
        stub_path = plugin_config_service.config_dir.joinpath("airflow.cfg")
        invoker = invoker_factory(
            project,
            self,
            prepare_with_session=session,
            plugin_config_service=plugin_config_service,
        )

        try:
            # generate the default `airflow.cfg`
            handle = invoker.invoke(
                "--help", stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            )
            handle.wait()
            logging.debug(f"Generated default '{str(airflow_cfg_path)}'")

            # move it to the config dir
            shutil.move(airflow_cfg_path, stub_path)
            airflow_cfg_path = stub_path
            logging.debug(f"Moved to '{str(stub_path)}'")

            # open the configuration and update it
            # now we let's update the config to use our stubs
            airflow_cfg = configparser.ConfigParser()

            with airflow_cfg_path.open() as cfg:
                airflow_cfg.read_file(cfg)
                logging.debug(f"Loaded '{str(airflow_cfg_path)}'")

            config = {}
            for key, value in plugin_settings_service.as_config(session, self).items():
                nest(config, key, str(value))

            for section, cfg in config.items():
                airflow_cfg[section].update(cfg)
                logging.debug(f"\tUpdated section [{section}] with {cfg}")

            with airflow_cfg_path.open("w") as cfg:
                airflow_cfg.write(cfg)
                logging.debug(f"Saved '{str(airflow_cfg_path)}'")

            # we've changed the configuration here, so we need to call
            # prepare again on the invoker so it re-reads the configuration
            # for the Airflow plugin
            invoker.prepare(session)
            handle = invoker.invoke(
                "initdb", stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            initdb = handle.wait()

            if initdb:
                raise SubprocessError("airflow initdb failed", handle)

            logging.debug(f"Completed `airflow initdb`")
        finally:
            session.close()



import asyncio
import logging
import sys
import os
from typing import Optional

from .config_service import ConfigService
from .db import project_engine
from .plugin import PluginType, PluginRef
from .plugin_invoker import invoker_factory
from .venv_service import VenvService
from .logging import capture_subprocess_output
from .elt_context import ELTContext


class DbtService:
    def __init__(self, project, config_service=None):
        self.project = project
        self.config_service = config_service or ConfigService(project)
        self.project_dir = self.project.root_dir("transform")
        self.profile_dir = self.project_dir.joinpath("profile")

        self._plugin = None

    @property
    def exec_path(self):
        return self.venv_service.exec_path("dbt", namespace=PluginType.TRANSFORMERS)

    @property
    def plugin(self):
        if not self._plugin:
            self._plugin = self.config_service.find_plugin("dbt")

        return self._plugin

    def project_args(self):
        return [
            "--project-dir",
            str(self.project_dir),
            "--profiles-dir",
            str(self.profile_dir),
            "--profile",
            "meltano",
        ]

    def dbt_invoker(self) -> "PluginInvoker":
        _, self._Session = project_engine(self.project)
        session = self._Session()

        try:
            return invoker_factory(
                self.project, self.plugin, prepare_with_session=session
            )
        finally:
            session.close()

    async def invoke(self, cmd, *args, **kwargs):
        invoker = self.dbt_invoker()
        handle = await invoker.invoke_async(
            cmd,
            *args,
            **kwargs,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        await asyncio.wait(
            [
                capture_subprocess_output(handle.stdout, sys.stdout),
                capture_subprocess_output(handle.stderr, sys.stderr),
                handle.wait(),
            ],
            return_when=asyncio.ALL_COMPLETED,
        )

        if handle.returncode:
            raise Exception(
                f"dbt {cmd} didn't exit cleanly. Exit code: {handle.returncode}"
            )

    async def deps(self):
        await self.invoke("deps")

    async def clean(self):
        await self.invoke("clean")

    async def compile(self, models=None, **kwargs):
        params = self.project_args()

        if models:
            # Always include the my_meltano_project model
            all_models = f"{models} my_meltano_project"
            params.extend(["--models", all_models])

        await self.invoke("compile", *params, **kwargs)

    async def run(self, models=None, **kwargs):
        params = self.project_args()

        if models:
            # Always include the my_meltano_project model
            all_models = f"{models} my_meltano_project"
            params.extend(["--models", all_models])

        await self.invoke("run", *params, **kwargs)



import os
import io
import yaml
import requests
import logging
import shutil
from copy import deepcopy
from typing import Dict, Iterator, Optional
from itertools import groupby, chain

import meltano.core.bundle as bundle
from .behavior.versioned import Versioned, IncompatibleVersionError
from .behavior.canonical import Canonical
from .config_service import ConfigService
from .plugin import Plugin, PluginInstall, PluginType, PluginRef
from .plugin.factory import plugin_factory


class PluginNotFoundError(Exception):
    pass


class DiscoveryInvalidError(Exception):
    """Occurs when the discovery.yml fails to be parsed."""

    pass


class DiscoveryUnavailableError(Exception):
    """Occurs when the discovery.yml cannot be found or downloaded."""

    pass


MELTANO_DISCOVERY_URL = "https://www.meltano.com/discovery.yml"

# Increment this version number whenever the schema of discovery.yml is changed.
# See https://www.meltano.com/docs/contributing.html#discovery-yml-version for more information.
VERSION = 10


class DiscoveryFile(Canonical):
    def __init__(self, **attrs):
        version = int(attrs.pop("version", 1))

        super().__init__(version=version)

        for plugin_type, plugin_defs in attrs.items():
            self[plugin_type] = []

            for plugin_def in plugin_defs:
                plugin = Plugin(
                    plugin_type,
                    plugin_def.pop("name"),
                    plugin_def.pop("namespace"),
                    **plugin_def,
                )
                self[plugin_type].append(plugin)


class PluginDiscoveryService(Versioned):
    __version__ = VERSION

    def __init__(
        self,
        project,
        config_service: ConfigService = None,
        discovery: Optional[Dict] = None,
    ):
        self.project = project
        self.config_service = config_service or ConfigService(project)
        self._discovery = DiscoveryFile.parse(discovery)

    @property
    def backend_version(self):
        return int(self._discovery.version)

    @property
    def discovery(self):
        """
        Return first compatible discovery manifest from these locations:

        - project local `discovery.yml`
        - http://meltano.com/discovery.yml
        - .meltano/cache/discovery.yml
        - meltano.core.bundle
        """
        if self._discovery:
            return self._discovery

        try:
            local_discovery = self.project.root.joinpath("discovery.yml")

            if local_discovery.is_file():
                with local_discovery.open() as local:
                    self._discovery = self.load_discovery(local)
            else:
                self._discovery = self.fetch_discovery()
        except DiscoveryInvalidError as e:
            # let's use the bundled one
            logging.error(
                "Meltano could not parse the `discovery.yml` returned by the server."
            )
            logging.debug(str(e))
            self._discovery = self.cached_discovery

        try:
            self.ensure_compatible()
        except IncompatibleVersionError as e:
            logging.fatal(
                "This version of Meltano cannot parse the plugins manifest, please update Meltano."
            )
            self._discovery = None
            raise

        return self._discovery

    def fetch_discovery(self):
        try:
            response = requests.get(MELTANO_DISCOVERY_URL)
            response.raise_for_status()
            self._discovery = self.load_discovery(io.StringIO(response.text))
            self.ensure_compatible()

            with self.cached_discovery_file.open("w") as discovery_cache:
                discovery_cache.write(response.text)
            logging.debug(f"Discovery cache updated at {self.cached_discovery_file}")
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
            yaml.YAMLError,
        ) as e:
            # let's try loading the cache instead
            self._discovery = self.cached_discovery
        except IncompatibleVersionError as e:
            # let's default to the local cache
            if e.backend_version > e.version:
                logging.warning("You must update Meltano to get new plugins updates.")
            self._discovery = self.cached_discovery

        if not self._discovery:
            raise DiscoveryInvalidError("Cannot parse discovery.yml.")

        return self._discovery

    @property
    def cached_discovery(self) -> DiscoveryFile:
        try:
            with self.cached_discovery_file.open() as discovery_cache:
                return self.load_discovery(discovery_cache)
        except (yaml.YAMLError, FileNotFoundError, DiscoveryInvalidError):
            logging.debug(
                f"Discovery cache corrupted or missing, falling back on the bundled discovery."
            )
            shutil.copy(bundle.find("discovery.yml"), self.cached_discovery_file)

            # load it back
            with self.cached_discovery_file.open() as discovery_cache:
                return self.load_discovery(discovery_cache)

    @property
    def cached_discovery_file(self):
        return self.project.meltano_dir("cache", "discovery.yml")

    def load_discovery(self, discovery_file) -> DiscoveryFile:
        try:
            return DiscoveryFile.parse(yaml.safe_load(discovery_file))
        except Exception as err:
            raise DiscoveryInvalidError(str(err))

    def plugins(self) -> Iterator[Plugin]:
        discovery_plugins = (
            plugin
            for plugin_type in PluginType
            for plugin in self.discovery[plugin_type]
        )

        yield from self.custom_plugins()
        yield from discovery_plugins

    def custom_plugins(self) -> Iterator[Plugin]:
        # some plugins in the Meltano file might be custom, thus they
        # serve both as `PluginInstall` and `Plugin`
        custom_plugins = (
            Plugin(custom_plugin.type, custom_plugin.name, **custom_plugin._extras)
            for custom_plugin in self.config_service.plugins()
            if custom_plugin.is_custom()
        )

        yield from custom_plugins

    def find_plugin(self, plugin_type: PluginType, plugin_name: str):
        name, _ = PluginRef.parse_name(plugin_name)
        try:
            return next(
                plugin
                for plugin in self.plugins()
                if (plugin.type == plugin_type and plugin.name == name)
            )
        except StopIteration:
            raise PluginNotFoundError(name)

    def discover(self, plugin_type: PluginType = None):
        """Return a pretty printed list of available plugins."""
        enabled_plugin_types = (plugin_type,) if plugin_type else iter(PluginType)

        return {
            plugin_type: [p.name for p in plugins]
            for plugin_type, plugins in groupby(self.plugins(), lambda p: p.type)
            if plugin_type in enabled_plugin_types
        }

    def list_discovery(self, discovery):
        return "\n".join(
            plugin.name for plugin in self.plugins() if plugin.type == discovery
        )



import datetime
import logging
import logging.handlers
import os
import atexit
from flask import Flask, request, g
from flask_login import current_user
from flask_cors import CORS
from urllib.parse import urlsplit

import meltano
from meltano.core.project import Project
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.compiler.project_compiler import ProjectCompiler
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.db import project_engine
from meltano.api.config import VERSION_HEADER

from werkzeug.wsgi import DispatcherMiddleware
from meltano.oauth.app import app as oauth_service


logger = logging.getLogger(__name__)


def create_app(config={}):
    project = Project.find()

    app = Flask(
        __name__, instance_path=str(project.root), instance_relative_config=True
    )

    app.config.from_object("meltano.api.config")
    app.config.from_pyfile("ui.cfg", silent=True)
    app.config.update(**config)

    if app.env == "production":
        from meltano.api.config import ensure_secure_setup

        app.config.from_object("meltano.api.config.Production")
        ensure_secure_setup(app)

    # register
    project_engine(
        project, engine_uri=app.config["SQLALCHEMY_DATABASE_URI"], default=True
    )

    # Initial compilation
    compiler = ProjectCompiler(project)
    try:
        compiler.compile()
    except Exception as e:
        pass

    # Logging
    file_handler = logging.handlers.RotatingFileHandler(
        str(project.run_dir("meltano-ui.log")), backupCount=3
    )
    stdout_handler = logging.StreamHandler()

    logger.setLevel(logging.DEBUG)
    logger.addHandler(file_handler)
    logger.addHandler(stdout_handler)

    # 1) Extensions
    security_options = {}

    from .models import db
    from .mail import mail
    from .executor import setup_executor
    from .security import security, users, setup_security
    from .security.oauth import setup_oauth
    from .json import setup_json

    db.init_app(app)
    mail.init_app(app)
    setup_executor(app, project)
    setup_security(app, project)
    setup_oauth(app)
    setup_json(app)

    # we need to setup CORS for development
    if app.env == "development":
        CORS(app, origins="http://localhost:8080", supports_credentials=True)

    # 2) Register the URL Converters
    from .url_converters import PluginRefConverter

    app.url_map.converters["plugin_ref"] = PluginRefConverter

    # 3) Register the controllers

    from .controllers.dashboards import dashboardsBP
    from .controllers.embeds import embedsBP
    from .controllers.reports import reportsBP
    from .controllers.repos import reposBP
    from .controllers.settings import settingsBP
    from .controllers.sql import sqlBP
    from .controllers.orchestrations import orchestrationsBP
    from .controllers.plugins import pluginsBP
    from .controllers.root import root, api_root

    app.register_blueprint(dashboardsBP)
    app.register_blueprint(embedsBP)
    app.register_blueprint(reportsBP)
    app.register_blueprint(reposBP)
    app.register_blueprint(settingsBP)
    app.register_blueprint(sqlBP)
    app.register_blueprint(orchestrationsBP)
    app.register_blueprint(pluginsBP)
    app.register_blueprint(root)
    app.register_blueprint(api_root)

    if app.config["PROFILE"]:
        from .profiler import init

        init(app)

    # Google Analytics setup
    tracker = GoogleAnalyticsTracker(project)

    @app.before_request
    def setup_js_context():
        # setup the appUrl
        appUrl = urlsplit(request.host_url)
        g.jsContext = {"appUrl": appUrl.geturl()[:-1]}

        if tracker.send_anonymous_usage_stats:
            g.jsContext["isSendAnonymousUsageStats"] = True
            g.jsContext["projectId"] = tracker.project_id

        g.jsContext["version"] = meltano.__version__

        # setup the oauthServiceUrl
        g.jsContext["oauthServiceUrl"] = app.config["MELTANO_OAUTH_SERVICE_URL"]

    @app.after_request
    def after_request(res):
        res.headers[VERSION_HEADER] = meltano.__version__
        return res

    # create the dispatcher to host the `OAuthService`
    app.wsgi_app = DispatcherMiddleware(app.wsgi_app, {"/-/oauth": oauth_service})

    return app



import os
import warnings
import meltano
import importlib

# disable the psycopg2 warning
# this needs to run before `psycopg2` is imported
warnings.filterwarnings("ignore", category=UserWarning, module="psycopg2")

workers = int(os.getenv("WORKERS", 4))
timeout = 600


def on_reload(arbiter):
    importlib.reload(meltano)
    print(f"Meltano version reloaded to {meltano.__version__}")



from typing import Dict

from meltano.core.behavior.hookable import hook
from meltano.core.connection_service import ConnectionService
from meltano.core.db import project_engine
from . import SingerPlugin, PluginType


class SingerTarget(SingerPlugin):
    __plugin_type__ = PluginType.LOADERS

    def exec_args(self, plugin_invoker):
        args = ["--config", plugin_invoker.files["config"]]

        return args

    @property
    def config_files(self):
        return {"config": "target.config.json"}

    @property
    def output_files(self):
        return {"state": "new_state.json"}



import json
from typing import Dict
from pathlib import Path


class M5ocFile:
    def __init__(self, path: Path, content: Dict):
        self.path = path
        self.content = content

    @classmethod
    def load(cls, file):
        return cls(file.name, json.load(file))



import logging

from .compiler.project_compiler import ProjectCompiler
from .plugin_discovery_service import PluginDiscoveryService
from .project_add_service import ProjectAddService
from .config_service import ConfigService
from .venv_service import VenvService
from .utils import noop
from .plugin import PluginType, PluginInstall, Plugin, PluginRef
from .project import Project
from .error import (
    PluginInstallError,
    PluginInstallWarning,
    SubprocessError,
    PluginNotInstallable,
)


def installer_factory(project, plugin, *args, **kwargs):
    cls = PipPluginInstaller

    if hasattr(plugin.__class__, "__installer_cls__"):
        cls = plugin.__class__.__installer_cls__

    return cls(project, plugin, *args, **kwargs)


class PluginInstallService:
    def __init__(self, project: Project, config_service: ConfigService = None):
        self.project = project
        self.config_service = config_service or ConfigService(project)

    def install_all_plugins(self, status_cb=noop):
        # TODO: config service returns PluginInstall, not Plugin
        return self.install_plugins(self.config_service.plugins())

    def install_plugins(self, plugins, status_cb=noop):
        errors = []
        installed = []
        has_model = False

        for plugin in plugins:
            status = {"plugin": plugin, "status": "running"}
            status_cb(status)

            try:
                self.install_plugin(plugin, compile_models=False)

                if plugin.type is PluginType.MODELS:
                    has_model = True

                status["status"] = "success"
                installed.append(status)
            except PluginInstallError as err:
                status["status"] = "error"
                status["message"] = str(err)
                status["details"] = err.process.stderr
                errors.append(status)
            except PluginInstallWarning as warn:
                status["status"] = "warning"
                status["message"] = str(warn)
                errors.append(status)
            except PluginNotInstallable as info:
                # let's totally ignore these plugins
                pass

            status_cb(status)

        if has_model:
            self.compile_models()

        return {"errors": errors, "installed": installed}

    def install_plugin(self, plugin: PluginInstall, compile_models=True):
        if not plugin.is_installable():
            raise PluginNotInstallable()

        with plugin.trigger_hooks("install", self.project):
            run = installer_factory(self.project, plugin).install()

            if compile_models and plugin.type is PluginType.MODELS:
                self.compile_models()

            return run

    def compile_models(self):
        compiler = ProjectCompiler(self.project)
        try:
            compiler.compile()
        except Exception:
            pass


class PipPluginInstaller:
    def __init__(self, project, plugin: PluginRef, venv_service: VenvService = None):
        self.plugin = plugin
        self.venv_service = venv_service or VenvService(project)

    def install(self):
        try:
            self.venv_service.create(namespace=self.plugin.type, name=self.plugin.name)
            return self.venv_service.install(
                namespace=self.plugin.type,
                name=self.plugin.name,
                pip_url=self.plugin.pip_url,
            )
        except SubprocessError as err:
            raise PluginInstallError(
                f"{self.plugin.name} has an installation issue. {err}", err.process
            )



import logging

from meltano.core.plugin import PluginInstall, PluginType
from meltano.core.behavior.hookable import hook
from meltano.core.m5o.m5o_collection_parser import (
    M5oCollectionParser,
    M5oCollectionParserTypes,
)
from meltano.core.venv_service import VirtualEnv
from meltano.core.m5o.reports_service import ReportAlreadyExistsError, ReportsService
from meltano.core.m5o.dashboards_service import (
    DashboardAlreadyExistsError,
    DashboardsService,
)


class DashboardPlugin(PluginInstall):
    __plugin_type__ = PluginType.DASHBOARDS

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)

    @hook("after_install")
    def after_install(self, project, args=[]):
        venv = VirtualEnv(project.plugin_dir(self, "venv"))
        packages_dir = venv.site_packages_dir

        logging.debug(f"Importing reports...")

        reports_parser = M5oCollectionParser(
            packages_dir.joinpath("reports"), M5oCollectionParserTypes.Report
        )
        reports = reports_parser.parse()
        logging.debug(f"Found {len(reports)} reports to import")

        reports_service = ReportsService(project)
        self.import_reports(reports, reports_service=reports_service)
        logging.debug(f"Finished importing {len(reports)} reports")

        logging.debug(f"Importing dashboards...")

        dashboards_parser = M5oCollectionParser(
            packages_dir.joinpath("dashboards"), M5oCollectionParserTypes.Dashboard
        )
        dashboards = dashboards_parser.parse()
        logging.debug(f"Found {len(dashboards)} dashboards to import")

        dashboards_service = DashboardsService(project)
        self.import_dashboards(dashboards, dashboards_service=dashboards_service)
        logging.debug(f"Finished importing {len(reports)} reports")

    def import_reports(self, reports, reports_service):
        for report in reports:
            self.import_report(report, reports_service=reports_service)

    def import_dashboards(self, dashboards, dashboards_service):
        for dashboard in dashboards:
            original_report_ids = dashboard["report_ids"]

            dashboard = self.import_dashboard(
                dashboard, dashboards_service=dashboards_service
            )

            self.add_reports_to_dashboard(
                dashboard, original_report_ids, dashboards_service=dashboards_service
            )

    def import_report(self, report, reports_service):
        report, existing = self.import_record(
            report,
            save_func=reports_service.save_report,
            already_exists_error=ReportAlreadyExistsError,
        )

        if existing:
            logging.debug(
                f"Found existing report with name '{report['name']}', ID '{report['id']}'"
            )
        else:
            logging.debug(
                f"Added report with name '{report['name']}', ID '{report['id']}'"
            )

        return report

    def import_dashboard(self, dashboard, dashboards_service):
        dashboard, existing = self.import_record(
            dashboard,
            save_func=dashboards_service.save_dashboard,
            already_exists_error=DashboardAlreadyExistsError,
        )

        if existing:
            logging.debug(
                f"Found existing dashboard with name '{dashboard['name']}', ID '{dashboard['id']}'"
            )
        else:
            logging.debug(
                f"Added dashboard with name '{dashboard['name']}', ID '{dashboard['id']}'"
            )

        return dashboard

    def import_record(self, record, save_func, already_exists_error):
        """
        Imports record.

        Returns `(record: Dict, existing: Bool)`
        """
        original_name = record["name"]
        counter = 1

        while True:
            try:
                record = save_func(record)
                return (record, False)
            except already_exists_error as e:
                if e.field == "slug":
                    counter += 1
                    record["name"] = f"{original_name} {counter}"

                    # Retry!
                    continue
                else:
                    record = e.record
                    return (record, True)

    def add_reports_to_dashboard(self, dashboard, report_ids, dashboards_service):
        for report_id in report_ids:
            dashboards_service.add_report_to_dashboard(
                {"dashboard_id": dashboard["id"], "report_id": report_id}
            )

            logging.debug(
                f"Added report with ID '{report_id}' to dashboard with ID '{dashboard['id']}'"
            )



# This file is managed by Meltano
# Do not edit this file directly

# If you want to define a custom DAG, create
# a new file under orchestrate/dags/ and Airflow
# will pick it up automatically.

import os
import logging
from airflow import DAG
from airflow.operators.bash_operator import BashOperator
from datetime import datetime, time, timedelta, MINYEAR

from meltano.core.schedule_service import ScheduleService
from meltano.core.project import Project
from meltano.core.utils import coerce_datetime


project = Project.find()
schedule_service = ScheduleService(project)

default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "email_on_failure": False,
    "email_on_retry": False,
    "catchup": False,
    "retries": 1,
    "retry_delay": timedelta(minutes=5),
    "concurrency": 1,
}

for schedule in schedule_service.schedules():
    args = default_args.copy()

    if schedule.start_date:
        args["start_date"] = coerce_datetime(schedule.start_date)

    if schedule.interval == "@once":
        logging.info(
            f"No DAG created for schedule '{schedule.name}' because it is set to `@once`."
        )
        continue

    dag_id = f"meltano_{schedule.name}"

    # from https://airflow.apache.org/docs/stable/scheduler.html#backfill-and-catchup
    #
    # It is crucial to set `catchup` to False so that Airflow only create a single job
    # at the tail end of date window we want to extract data.
    #
    # Because our extractors do not support date-window extraction, it serves no
    # purpose to enqueue date-chunked jobs for complete extraction window.
    dag = DAG(
        dag_id, catchup=False, default_args=args, schedule_interval=schedule.interval
    )

    elt = BashOperator(
        task_id="extract_load",
        bash_command=f"echo $PATH; echo $VIRTUAL_ENV; cd {str(project.root)}; .meltano/run/bin elt {schedule.extractor} {schedule.loader} --job_id={schedule.name} --transform={schedule.transform}",
        dag=dag,
        env={
            # inherit the current env
            **os.environ,
            **schedule.env,
        },
    )

    # register the dag
    globals()[dag_id] = dag



"""ensure user.login_count default value

Revision ID: 990c0665f3ce
Revises: a3e2b0a4937d
Create Date: 2020-01-15 15:58:25.416129

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import table, column


# revision identifiers, used by Alembic.
revision = "990c0665f3ce"
down_revision = "a3e2b0a4937d"
branch_labels = None
depends_on = None


def upgrade():
    # update the existing Users
    user = table("user", column("login_count", sa.Integer))
    connection = op.get_bind()

    connection.execute(
        user.update().where(user.c.login_count == None).values({"login_count": 0})
    )


def downgrade():
    pass



import os
import yaml
import logging
from typing import Dict, List, Optional, Iterable

from meltano.core.utils import nest, NotFound
from .project import Project
from .plugin import Plugin, PluginInstall, PluginType, PluginRef, Profile
from .plugin.factory import plugin_factory
from .plugin.error import PluginMissingError


class ConfigService:
    def __init__(self, project: Project):
        self.project = project

    def make_meltano_secret_dir(self):
        os.makedirs(self.project.meltano_dir(), exist_ok=True)

    def add_to_file(self, plugin: PluginInstall):
        with self.project.meltano_update() as meltano_yml:
            if not plugin in self.plugins():
                if not plugin.type in meltano_yml.plugins:
                    meltano_yml.plugins[plugin.type] = []

                meltano_yml.plugins[plugin.type].append(plugin)
            else:
                logging.warning(
                    f"{plugin.name} is already present, use `meltano install` to install it."
                )

        return plugin_factory(plugin.type, plugin.canonical())

    def has_plugin(self, plugin_name: str):
        try:
            self.find_plugin(plugin_name)
            return True
        except PluginMissingError:
            return False

    def find_plugin(self, plugin_name: str, plugin_type: Optional[PluginType] = None):
        name, profile_name = PluginRef.parse_name(plugin_name)
        try:
            plugin = next(
                plugin
                for plugin in self.plugins()
                if (
                    plugin.name == name
                    and (plugin_type is None or plugin.type == plugin_type)
                )
            )
            plugin.use_profile(profile_name)

            return plugin
        except StopIteration as stop:
            raise PluginMissingError(name) from stop

    def get_plugin(self, plugin_ref: PluginRef) -> PluginInstall:
        try:
            plugin = next(plugin for plugin in self.plugins() if plugin == plugin_ref)
            plugin.use_profile(plugin_ref.current_profile_name)

            return plugin
        except StopIteration as stop:
            raise PluginMissingError(plugin_ref.name) from stop

    def get_extractors(self):
        return filter(lambda p: p.type == PluginType.EXTRACTORS, self.plugins())

    def get_loaders(self):
        return filter(lambda p: p.type == PluginType.LOADERS, self.plugins())

    def get_transforms(self):
        return filter(lambda p: p.type == PluginType.TRANSFORMS, self.plugins())

    def get_models(self):
        return filter(lambda p: p.type == PluginType.MODELS, self.plugins())

    def get_dashboards(self):
        return filter(lambda p: p.type == PluginType.DASHBOARDS, self.plugins())

    def get_transformers(self):
        return filter(lambda p: p.type == PluginType.TRANSFORMERS, self.plugins())

    def update_plugin(self, plugin: PluginInstall):
        with self.project.meltano_update() as meltano:
            # find the proper plugin to update
            idx, outdated = next(
                (i, it)
                for i, it in enumerate(meltano.plugins[plugin.type])
                if it == plugin
            )

            meltano.plugins[plugin.type][idx] = plugin

            return outdated

    def plugins(self) -> Iterable[PluginInstall]:
        for plugin_type, plugins in self.project.meltano.plugins:
            yield from plugins



import os
import sqlalchemy
import logging
from typing import Iterable, Dict, Tuple, List
from copy import deepcopy
from enum import Enum

from meltano.core.utils import nest, flatten, find_named, NotFound, truthy
from meltano.core.config_service import ConfigService
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.error import Error
from . import PluginRef, PluginType, Plugin, PluginInstall, SettingDefinition, Profile
from .error import PluginMissingError
from .setting import PluginSetting, REDACTED_VALUE


class PluginSettingMissingError(Error):
    """Occurs when a setting is missing."""

    def __init__(self, plugin: PluginRef, name: str):
        super().__init__(f"Cannot find setting {name} in {plugin}")


class PluginSettingValueSource(str, Enum):
    ENV = "env"  # 0
    MELTANO_YML = "meltano_yml"  # 1
    DB = "db"  # 2
    DEFAULT = "default"  # 3


class PluginSettingsService:
    def __init__(
        self,
        project,
        config_service: ConfigService = None,
        plugin_discovery_service: PluginDiscoveryService = None,
        show_hidden=True,
    ):
        self.project = project
        self.config_service = config_service or ConfigService(project)
        self.discovery_service = plugin_discovery_service or PluginDiscoveryService(
            project
        )
        self.show_hidden = show_hidden

    @classmethod
    def unredact(cls, values: dict) -> Dict:
        """
        Removes any redacted values in a dictionary.
        """

        return {k: v for k, v in values.items() if v != REDACTED_VALUE}

    @classmethod
    def is_kind_redacted(cls, kind) -> bool:
        return kind in ("password", "oauth")

    def profile_with_config(
        self, session, plugin: PluginRef, profile: Profile, redacted=False
    ):
        plugin_install = self.get_install(plugin)

        plugin_install.use_profile(profile)

        full_config = self.config_with_sources(
            session, plugin_install, redacted=redacted
        )

        return {
            **profile.canonical(),
            "config": {key: config["value"] for key, config in full_config.items()},
            "config_sources": {
                key: config["source"] for key, config in full_config.items()
            },
        }

    def profiles_with_config(
        self, session, plugin: PluginRef, redacted=False
    ) -> List[Dict]:
        plugin_install = self.get_install(plugin)

        return [
            self.profile_with_config(session, plugin, profile, redacted=redacted)
            for profile in (Profile.DEFAULT, *plugin_install.profiles)
        ]

    def config_with_sources(
        self,
        session,
        plugin: PluginRef,
        sources: List[PluginSettingValueSource] = None,
        redacted=False,
    ):
        # defaults to the meltano.yml for extraneous settings
        plugin_install = self.get_install(plugin)
        config = {
            key: {"value": value, "source": PluginSettingValueSource.MELTANO_YML}
            for key, value in deepcopy(plugin_install.current_config).items()
        }

        # definition settings
        for setting in self.definitions(plugin):
            value, source = self.get_value(session, plugin, setting.name)
            if sources and source not in sources:
                continue

            # we don't want to leak secure informations
            # so we redact all `passwords`
            if redacted and value and self.is_kind_redacted(setting.kind):
                value = REDACTED_VALUE

            config[setting.name] = {"value": value, "source": source}

        return config

    def as_config(self, *args, **kwargs) -> Dict:
        full_config = self.config_with_sources(*args, **kwargs)

        return {key: config["value"] for key, config in full_config.items()}

    def as_env(
        self, session, plugin: PluginRef, sources: List[PluginSettingValueSource] = None
    ) -> Dict[str, str]:
        # defaults to the meltano.yml for extraneous settings
        plugin_def = self.get_definition(plugin)
        env = {}

        for setting in self.definitions(plugin):
            value, source = self.get_value(session, plugin, setting.name)
            if sources and source not in sources:
                logging.debug(f"Setting {setting.name} is not in sources: {sources}.")
                continue

            env_key = self.setting_env(setting, plugin_def)
            env[env_key] = str(value)

        return env

    def set(self, session, plugin: PluginRef, name: str, value, enabled=True):
        try:
            plugin_def = self.get_definition(plugin)
            setting_def = self.find_setting(plugin, name)
            env_key = self.setting_env(setting_def, plugin_def)

            if env_key in os.environ:
                logging.warning(f"Setting `{name}` is currently set via ${env_key}.")
                return

            if value == REDACTED_VALUE:
                return

            setting = PluginSetting(
                namespace=plugin.qualified_name, name=name, value=value, enabled=enabled
            )

            session.merge(setting)
            session.commit()

            return setting
        except StopIteration:
            logging.warning(f"Setting `{name}` not found.")

    def unset(self, session, plugin: PluginRef, name: str):
        session.query(PluginSetting).filter_by(
            namespace=plugin.qualified_name, name=name
        ).delete()
        session.commit()

    def get_definition(self, plugin: PluginRef) -> Plugin:
        return self.discovery_service.find_plugin(plugin.type, plugin.name)

    def find_setting(self, plugin: PluginRef, name: str) -> Dict:
        try:
            return find_named(self.definitions(plugin), name)
        except NotFound as err:
            raise PluginSettingMissingError(plugin, name) from err

    def definitions(self, plugin_ref: PluginRef) -> Iterable[Dict]:
        settings = set()
        plugin_def = self.get_definition(plugin_ref)
        only_visible = lambda s: s.kind != "hidden" or self.show_hidden

        return list(filter(only_visible, plugin_def.settings))

    def get_install(self, plugin: PluginRef) -> PluginInstall:
        return self.config_service.get_plugin(plugin)

    def setting_env(self, setting_def, plugin_def):
        if setting_def.env:
            return setting_def.env

        parts = (plugin_def.namespace, setting_def["name"])
        process = lambda s: s.replace(".", "__").upper()
        return "_".join(map(process, parts))

    def get_value(self, session, plugin: PluginRef, name: str):
        plugin_install = self.get_install(plugin)
        plugin_def = self.get_definition(plugin)
        setting_def = self.find_setting(plugin, name)

        try:
            env_key = self.setting_env(setting_def, plugin_def)

            # priority 1: environment variable
            if env_key in os.environ:
                logging.debug(
                    f"Found ENV variable {env_key} for {plugin_def.namespace}:{name}"
                )

                value = os.environ[env_key]
                if setting_def.kind == "boolean":
                    value = truthy(value)

                return (value, PluginSettingValueSource.ENV)

            # priority 2: installed configuration
            if setting_def.name in plugin_install.current_config:
                return (
                    plugin_install.current_config[setting_def.name],
                    PluginSettingValueSource.MELTANO_YML,
                )

            # priority 3: settings database
            return (
                (
                    session.query(PluginSetting)
                    .filter_by(namespace=plugin.qualified_name, name=name, enabled=True)
                    .one()
                    .value
                ),
                PluginSettingValueSource.DB,
            )
        except sqlalchemy.orm.exc.NoResultFound:
            # priority 4: setting default value
            # that means it was not overriden
            return setting_def.value, PluginSettingValueSource.DEFAULT



import logging
import threading
import time
import psutil

from meltano.core.project import Project
from meltano.core.plugin import PluginInstall, PluginType
from meltano.core.project_add_service import ProjectAddService
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.config_service import ConfigService, PluginMissingError
from meltano.core.plugin_invoker import invoker_factory
from meltano.core.db import project_engine
from meltano.core.utils.pidfile import PIDFile, UnknownProcessError


class AirflowWorker(threading.Thread):
    def __init__(self, project: Project):
        super().__init__(name="AirflowWorker")

        self.project = project
        self.config_service = ConfigService(project)
        self.add_service = ProjectAddService(
            project, config_service=self.config_service
        )
        self.install_service = PluginInstallService(
            project, config_service=self.config_service
        )
        self._plugin = None
        self.pid_file = PIDFile(self.project.run_dir("airflow", "scheduler.pid"))

    def kill_stale_workers(self):
        process = None
        try:
            process = self.pid_file.process
        except UnknownProcessError:
            pass

        if process is not None:
            logging.debug(
                f"Process {process} is running, possibly stale, terminating it."
            )
            process.terminate()

            def on_terminate(process):
                logging.info(
                    f"Process {process} ended with exit code {process.returncode}"
                )

            _gone, alive = psutil.wait_procs(
                [process], timeout=5, callback=on_terminate
            )

            # kill the rest
            for process in alive:
                process.kill()

        try:
            self.pid_file.unlink()
        except:
            pass

    def start_all(self):
        _, Session = project_engine(self.project)
        logs_path = self.project.run_dir("airflow", "logs", "scheduler.log")

        try:
            session = Session()
            invoker = invoker_factory(
                self.project, self._plugin, prepare_with_session=session
            )

            with logs_path.open("w") as logs_file:
                scheduler = invoker.invoke(
                    "scheduler", "--pid", str(self.pid_file), stdout=logs_file
                )
                self.pid_file.write_pid(scheduler.pid)
        finally:
            session.close()

    def run(self):
        try:
            self._plugin = self.config_service.find_plugin("airflow")
        except PluginMissingError as err:
            self._plugin = self.add_service.add(PluginType.ORCHESTRATORS, "airflow")
            self.install_service.install_plugin(self._plugin)

        self.kill_stale_workers()
        self.start_all()

    def stop(self):
        self.kill_stale_workers()



import os
from secrets import token_hex

# session are used internally to verify the tokens
SECRET_KEY = token_hex(64)
FACEBOOK_CLIENT_ID = os.getenv("OAUTH_FACEBOOK_CLIENT_ID")
FACEBOOK_CLIENT_SECRET = os.getenv("OAUTH_FACEBOOK_CLIENT_SECRET")



import logging

from flask import Blueprint
from flask_login import current_user
from meltano.api.security import api_auth_required


VERSION = 1


class APIBlueprint(Blueprint):
    @property
    def api_version():
        return f"v{VERSION}"

    @classmethod
    def url_prefix(cls, name, version=VERSION):
        return f"/api/v{version}/{name}"

    def __init__(self, name, *args, **kwargs):
        # make sure the `url_prefix` is set
        url_prefix = kwargs.pop("url_prefix", self.url_prefix(name))

        super().__init__(name, *args, url_prefix=url_prefix, **kwargs)

        self.before_request(self.__class__.auth_filter)

    @api_auth_required
    def auth_filter():
        logging.debug(f"Authenticated as {current_user}")



import json
import os
import logging

from meltano.core.behavior.hookable import hook
from meltano.core.project import Project
from meltano.core.plugin import PluginInstall
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.db import project_engine


class SingerPlugin(PluginInstall):
    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)

    @hook("before_configure")
    def install_config_stub(self, invoker):
        project = invoker.project
        plugin_dir = project.plugin_dir(self)

        with open(plugin_dir.joinpath(self.config_files["config"]), "w") as config_stub:
            json.dump(invoker.plugin_config, config_stub)
            logging.debug(f"Created configuration stub at {config_stub}")



import datetime
import subprocess
import logging
import os
from functools import partial
from flask_executor import Executor

from meltano.core.plugin import PluginRef, PluginType
from meltano.core.project import Project


executor = Executor()


def setup_executor(app, project):
    executor.init_app(app)


def run_elt(project: Project, schedule_payload: dict):
    job_id = schedule_payload["name"]
    extractor = schedule_payload["extractor"]
    loader = schedule_payload["loader"]
    transform = schedule_payload.get("transform")

    cmd = [
        "meltano",
        "elt",
        "--job_id",
        job_id,
        extractor,
        loader,
        "--transform",
        transform,
    ]
    executor.submit(
        subprocess.run, cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
    )
    logging.debug(f"Defered `{' '.join(cmd)}` to the executor.")

    return job_id


def upgrade():
    cmd = ["meltano", "upgrade"]
    executor.submit(
        subprocess.run, cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
    )



import os
import json

from meltano.core.utils import slugify

from .m5o_collection_parser import M5oCollectionParser, M5oCollectionParserTypes
from .m5o_file_parser import MeltanoAnalysisFileParser


class ReportAlreadyExistsError(Exception):
    """Occurs when a report already exists."""

    def __init__(self, report, field):
        self.report = report
        self.field = field

    @property
    def record(self):
        return self.report


class ReportsService:
    VERSION = "1.0.0"

    def __init__(self, project):
        self.project = project

    def get_reports(self):
        reportsParser = M5oCollectionParser(
            self.project.analyze_dir("reports"), M5oCollectionParserTypes.Report
        )

        return reportsParser.parse()

    def get_report(self, report_id):
        reports = self.get_reports()
        report = next(filter(lambda r: r["id"] == report_id, reports), None)
        return report

    def save_report(self, data):
        if "id" in data:
            existing_report = self.get_report(data["id"])
            if existing_report:
                raise ReportAlreadyExistsError(existing_report, "id")

        name = data["name"]
        slug = slugify(name)
        file_path = self.project.analyze_dir("reports", f"{slug}.report.m5o")

        if os.path.exists(file_path):
            with file_path.open() as f:
                existing_report = json.load(f)
            raise ReportAlreadyExistsError(existing_report, "slug")

        data = MeltanoAnalysisFileParser.fill_base_m5o_dict(
            file_path.relative_to(self.project.root), slug, data
        )
        data["version"] = ReportsService.VERSION

        with file_path.open("w") as f:
            json.dump(data, f)

        return data

    def update_report(self, data):
        id = data["id"]
        existing_report = self.get_report(id)
        slug = existing_report["slug"]

        file_path = self.project.analyze_dir("reports", f"{slug}.report.m5o")

        new_name = data["name"]
        new_slug = slugify(new_name)
        new_file_path = self.project.analyze_dir("reports", f"{new_slug}.report.m5o")
        is_same_file = new_slug == slug
        if not is_same_file and os.path.exists(new_file_path):
            with new_file_path.open() as f:
                existing_report = json.load(f)
            raise ReportAlreadyExistsError(existing_report)

        os.remove(file_path)

        data["slug"] = new_slug
        data["path"] = str(new_file_path.relative_to(self.project.root))
        with new_file_path.open("w") as f:
            json.dump(data, f)

        return data



from meltano.core.project import Project
from meltano.core.schedule_service import ScheduleService

from .sql_helper import SqlHelper


class ReportsHelper:
    VERSION = "1.0.0"

    def get_report_with_query_results(self, report):
        project = Project.find()
        schedule_service = ScheduleService(project)
        sql_helper = SqlHelper()
        report_with_query_results = self.update_report_with_query_results(
            report, schedule_service, sql_helper
        )

        return report_with_query_results

    def update_report_with_query_results(self, report, schedule_service, sql_helper):
        m5oc = sql_helper.get_m5oc_topic(report["namespace"], report["model"])
        design = m5oc.design(report["design"])
        schedule = schedule_service.find_namespace_schedule(
            m5oc.content["plugin_namespace"]
        )

        sql_dict = sql_helper.get_sql(design, report["query_payload"])
        outgoing_sql = sql_dict["sql"]
        aggregates = sql_dict["aggregates"]

        report["query_results"] = sql_helper.get_query_results(
            schedule.loader, outgoing_sql
        )
        report["query_result_aggregates"] = aggregates

        return report



import yaml
import fnmatch
import copy
from typing import Dict, Union
from collections import namedtuple
from enum import Enum
from typing import Optional, Iterable

from meltano.core.behavior.hookable import HookObject
from meltano.core.behavior.canonical import Canonical
from meltano.core.behavior import NameEq
from meltano.core.utils import compact, find_named, NotFound


class YAMLEnum(str, Enum):
    def __str__(self):
        return self.value

    @staticmethod
    def yaml_representer(dumper, obj):
        return dumper.represent_scalar("tag:yaml.org,2002:str", str(obj))


yaml.add_multi_representer(YAMLEnum, YAMLEnum.yaml_representer)


class Profile(NameEq, Canonical):
    def __init__(self, name: str = None, label: str = None, config={}):
        self.name = name
        self.label = label
        self.config = config


Profile.DEFAULT = Profile(name="default", label="Default")


class SettingDefinition(NameEq, Canonical):
    def __init__(
        self,
        name: str = None,
        env: str = None,
        kind: str = None,
        value=None,
        label: str = None,
        documentation: str = None,
        description: str = None,
        tooltip: str = None,
        options: list = [],
        oauth: dict = {},
        placeholder: str = None,
        protected=False,
        **attrs,
    ):
        super().__init__(
            name=name,
            env=env,
            kind=kind,
            value=value,
            label=label,
            documentation=documentation,
            description=description,
            tooltip=tooltip,
            options=options,
            oauth=oauth,
            placeholder=placeholder,
            protected=protected,
            **attrs,
        )


class PluginType(YAMLEnum):
    EXTRACTORS = "extractors"
    LOADERS = "loaders"
    TRANSFORMS = "transforms"
    MODELS = "models"
    DASHBOARDS = "dashboards"
    ORCHESTRATORS = "orchestrators"
    TRANSFORMERS = "transformers"

    def __str__(self):
        return self.value

    @property
    def cli_command(self):
        """Makes it singular for `meltano add PLUGIN_TYPE`"""
        return self.value[:-1]

    @classmethod
    def value_exists(cls, value):
        return value in cls._value2member_map_


class PluginRef:
    def __init__(self, plugin_type: Union[str, PluginType], name: str):
        self._type = (
            plugin_type
            if isinstance(plugin_type, PluginType)
            else PluginType(plugin_type)
        )
        self.name, self._current_profile_name = self.parse_name(name)

    @classmethod
    def parse_name(cls, name: str):
        name, *profile_name = name.split("@")
        profile_name = next(iter(profile_name), Profile.DEFAULT.name)

        return (name, profile_name)

    @property
    def type(self):
        return self._type

    @property
    def current_profile_name(self):
        return self._current_profile_name

    @property
    def full_name(self):
        return f"{self.name}@{self.current_profile_name}"

    @property
    def qualified_name(self):
        parts = (self.type, self.name, self.current_profile_name)

        return ".".join(compact(parts))

    def __eq__(self, other):
        return self.name == other.name and self.type == other.type

    def __hash__(self):
        return hash((self.type, self.name, self.current_profile_name))


class PluginInstall(HookObject, Canonical, PluginRef):
    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        pip_url: Optional[str] = None,
        select=set(),
        config={},
        profiles=set(),
        executable: str = None,
        **attrs,
    ):
        super().__init__(plugin_type, name, **attrs)

        self.config = copy.deepcopy(config)
        self.profiles = set(map(Profile.parse, profiles))
        self.select = select
        self.pip_url = pip_url
        self.executable = executable
        self._extras = attrs

    def is_installable(self):
        return self.pip_url is not None

    def is_custom(self):
        try:
            return bool(self.namespace)
        except AttributeError:
            return False

    def get_profile(self, profile_name: str) -> Profile:
        if profile_name == Profile.DEFAULT.name:
            return Profile.DEFAULT

        return find_named(self.profiles, profile_name)

    def use_profile(self, profile_or_name: Union[str, Profile]):
        if profile_or_name is None:
            profile = Profile.DEFAULT
        elif isinstance(profile_or_name, Profile):
            profile = self.get_profile(profile_or_name.name)
        else:
            profile = self.get_profile(profile_or_name)

        self._current_profile_name = profile.name

    @property
    def current_profile(self):
        return self.get_profile(self.current_profile_name)

    @property
    def current_config(self):
        if self.current_profile is Profile.DEFAULT:
            return self.config

        return self.current_profile.config

    def exec_args(self, files: Dict):
        return []

    @property
    def config_files(self):
        """Return a list of stubbed files created for this plugin."""
        return dict()

    @property
    def output_files(self):
        return dict()

    def add_select_filter(self, filter: str):
        self.select.add(filter)

    def add_profile(self, name: str, config: dict = None, label: str = None):
        profile = Profile(name=name, config=config, label=label)

        self.profiles.add(profile)
        return profile


class Plugin(Canonical, PluginRef):
    """
    Args:
    name: The unique name for the installed plugin
    pip_url: The pip-compatible installation URI, like `git+https://…` or `-e /path/to/pkg`
    executable: The plugin executable name (default: <name>)
    """

    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        namespace: str,
        pip_url: Optional[str] = None,
        settings: list = [],
        settings_group_validation: list = [],
        docs=None,
        description=None,
        capabilities=set(),
        select=set(),
        **attrs,
    ):
        super().__init__(plugin_type, name, **attrs)

        self.namespace = namespace
        self.pip_url = pip_url
        self.settings = list(map(SettingDefinition.parse, settings))
        self.settings_group_validation = list(settings_group_validation)
        self.docs = docs
        self.description = description
        self.capabilities = set(capabilities)
        self.select = set(select)
        self._extras = attrs

    def as_installed(self) -> PluginInstall:
        return PluginInstall(
            self.type,
            self.name,
            pip_url=self.pip_url,
            select=self.select,
            **self._extras,
        )



from flask import Blueprint, url_for, redirect, render_template, jsonify
from authlib.flask.client import OAuth as OAuthClient


OAuth = OAuthClient()


def facebook(app):
    OAuth.register(
        "facebook",
        access_token_url="https://graph.facebook.com/v5.0/oauth/access_token",
        client_kwargs={"scope": "ads_read ads_management manage_pages"},
        authorize_url="https://www.facebook.com/v5.0/dialog/oauth",
    )

    oauthBP = Blueprint("OAuth.Facebook", __name__, url_prefix="/facebook")

    @oauthBP.route("/")
    def login():
        redirect_uri = url_for(".authorize", _external=True)

        return OAuth.facebook.authorize_redirect(redirect_uri, auth_type="rerequest")

    @oauthBP.route("/authorize")
    def authorize():
        token = OAuth.facebook.authorize_access_token()

        return render_template("token.html", token=token)

    app.register_blueprint(oauthBP)



import base64
import re
import sys
import flatten_dict
import os
import functools
from datetime import datetime, date, time
from copy import deepcopy
from typing import Union, Dict, Callable, Optional, Iterable
from requests.auth import HTTPBasicAuth
from pathlib import Path


TRUTHY = ("true", "1", "yes", "on")


class NotFound(Exception):
    """Occurs when an element is not found."""

    def __init__(self, name):
        super().__init__(f"{name} was not found.")


# from https://github.com/jonathanj/compose/blob/master/compose.py
def compose(*fs):
    """
    Create a function composition.
    :type *fs: ``iterable`` of 1-argument ``callable``s
    :param *fs: Iterable of 1-argument functions to compose, functions will be
        applied from last to first, in other words ``compose(f, g)(x) ==
        f(g(x))``.
    :return: I{callable} taking 1 argument.
    """
    return functools.reduce(lambda f, g: lambda x: f(g(x)), compact(fs), lambda x: x)


# from http://www.dolphmathews.com/2012/09/slugify-string-in-python.html
def slugify(s):
    """
    Simplifies ugly strings into something URL-friendly.
    >>> slugify("[Some] _ Article's Title--")
    'some-articles-title'
    """

    # "[Some] _ Article's Title--"
    # "[some] _ article's title--"
    s = s.lower()

    # "[some] _ article's_title--"
    # "[some]___article's_title__"
    for c in [" ", "-", ".", "/"]:
        s = s.replace(c, "_")

    # "[some]___article's_title__"
    # "some___articles_title__"
    s = re.sub("\W", "", s)

    # "some___articles_title__"
    # "some   articles title  "
    s = s.replace("_", " ")

    # "some   articles title  "
    # "some articles title "
    s = re.sub("\s+", " ", s)

    # "some articles title "
    # "some articles title"
    s = s.strip()

    # "some articles title"
    # "some-articles-title"
    s = s.replace(" ", "-")

    return s


def get_basic_auth(user, token):
    return HTTPBasicAuth(user, token)


def pop_all(keys, d: dict):
    return dict(map(lambda k: (k, d.pop(k)), keys))


def get_all(keys, d: dict, default=None):
    return dict(map(lambda k: (k, d.get(k, default)), keys))


def nest(d: dict, path: str, value={}):
    """
    Create a hierarchical dictionary path and return the leaf dict.

    >>> d = dict()
    >>> test = nest(d, "foo.bar.test")
    >>> test
    {}
    >>> d
    {'foo': {'bar': {'test': {}}}}
    >>> test["a"] = 1
    >>> d
    {'foo': {'bar': {'test': {'a': 1}}}}
    >>> alist = nest(d, "foo.bar.list", value=[])
    >>> alist.append("works")
    >>> d
    {'foo': {'bar': {'test': {'a': 1}}, 'list': ["works"]}}
    """
    cursor = d

    *initial, tail = path.split(".")

    # create the list of dicts
    for key in initial:
        if key not in cursor:
            cursor[key] = {}

        cursor = cursor[key]

    # We need to copy the value to make sure
    # the `value` parameter is not mutated.
    cursor[tail] = cursor.get(tail, deepcopy(value))

    return cursor[tail]


def flatten(d: Dict, reducer: Union[str, Callable] = "tuple", **kwargs):
    """Wrapper arround `flatten_dict.flatten` that adds `dot` reducer."""

    def dot_reducer(*xs):
        if xs[0] is None:
            return xs[1]
        else:
            return ".".join(xs)

    if reducer == "dot":
        reducer = dot_reducer

    return flatten_dict.flatten(d, reducer, **kwargs)


def compact(xs: Iterable) -> Iterable:
    return (x for x in xs if x is not None)


def file_has_data(file: Union[Path, str]):
    file = Path(file)  # ensure it is a Path object
    return file.exists() and file.stat().st_size > 0


def identity(x):
    return x


def noop(*_args, **_kwargs):
    pass


def map_dict(f: Callable, d: Dict):
    for k, v in d.items():
        yield k, f(v)


def truthy(val: str) -> bool:
    return str(val).lower() in TRUTHY


def coerce_datetime(d: Union[date, datetime]) -> Optional[datetime]:
    """Adds a `time` component to `d` if such a component is missing."""
    if d is None:
        return None

    if isinstance(d, datetime):
        return d

    return datetime.combine(d, time())


def iso8601_datetime(d: str) -> Optional[datetime]:
    if d is None:
        return None

    isoformats = [
        "%Y-%m-%dT%H:%M:%SZ",
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%d %H:%M:%S",
        "%Y-%m-%d",
    ]

    for format in isoformats:
        try:
            return coerce_datetime(datetime.strptime(d, format))
        except ValueError:
            pass

    raise ValueError(f"{d} is not a valid UTC date.")


def find_named(xs: Iterable[dict], name: str):
    try:
        return next(x for x in xs if x["name"] == name)
    except StopIteration as stop:
        raise NotFound(name)


def makedirs(func):
    @functools.wraps(func)
    def decorate(*args, **kwargs):
        path = func(*args, **kwargs)

        # if there is an extension, only create the base dir
        _, ext = os.path.splitext(path)
        if ext:
            dir = os.path.dirname(path)
        else:
            dir = path

        os.makedirs(dir, exist_ok=True)
        return path

    return decorate



import json
import base64
import gitlab
from authlib.flask.client import OAuth as OAuthClient
from flask import Blueprint, url_for, redirect, current_app, jsonify
from flask_security import current_user, AnonymousUser
from flask_security.utils import login_user, do_flash, url_for_security
from sqlalchemy.orm import joinedload
from sqlalchemy.orm.exc import NoResultFound

from .identity import users, FreeUser
from meltano.api.models.oauth import db, OAuth
from meltano.core.utils import compose


class OAuthError(Exception):
    pass


def base64_pad(s):
    padding = 4 - (len(s) % 4)
    return s + ("=" * padding)


jwt_decode = compose(json.loads, base64.urlsafe_b64decode, base64_pad)


def setup_oauth_gitlab(oauth):
    oauth.register(
        "gitlab",
        access_token_url="https://gitlab.com/oauth/token",
        client_kwargs={"scope": "openid read_user"},
        authorize_url="https://gitlab.com/oauth/authorize",
    )

    oauthBP = Blueprint("OAuth.GitLab", __name__, url_prefix="/oauth/gitlab")

    @oauthBP.route("/login")
    def login():
        redirect_uri = url_for(".authorize", _external=True)
        return oauth.gitlab.authorize_redirect(redirect_uri)

    @oauthBP.route("/authorize")
    def authorize():
        token = oauth.gitlab.authorize_access_token()

        try:
            identity = gitlab_token_identity(token)
            login_user(identity.user, remember=False)

            return redirect(url_for("root.default"))
        except OAuthError as e:
            do_flash(str(e))
            return redirect(url_for_security("login"))

    oauth.app.register_blueprint(oauthBP)


def setup_oauth(app):
    oauth = OAuthClient(app)
    setup_oauth_gitlab(oauth)


def gitlab_token_identity(token):
    # TODO: having to do another GET to grab the user info
    #       is subpar, but the returned JWT lacks the user's email
    #
    # It seems like validating the JWT token is not
    # currently working as the GitLab JWKs endpoint is unstable
    # See https://gitlab.com/gitlab-com/support-forum/issues/3666
    client = gitlab.Gitlab("http://gitlab.com", oauth_token=token["access_token"])
    client.auth()

    if client.user.state != "active":
        raise OAuthError("This account is not active.")

    token_user_email = users.get_user(client.user.email)
    token_user_username = users.get_user(client.user.username)

    if (
        token_user_email
        and token_user_username
        and token_user_email != token_user_username
    ):
        raise OAuthError(
            "This identity is already claimed by another user, please login."
        )
    else:
        token_user = token_user_email or token_user_username

    identity = None
    try:
        identity = (
            db.session.query(OAuth)
            .options(joinedload(OAuth.user))
            .filter_by(provider_id="gitlab", provider_user_id=client.user.id)
            .one()
        )
    except NoResultFound:
        identity = OAuth.from_token("gitlab", client.user.id, token)

    if current_user.is_authenticated and not isinstance(
        current_user._get_current_object(), FreeUser
    ):
        # map the identity to the current_user
        identity.user = current_user
        db.session.add(identity)
    elif not token_user:
        # no user has claimed this email yet
        # reserve it
        token_user = users.create_user(
            username=client.user.username, email=client.user.email
        )
        identity.user = token_user
        db.session.add(identity)
    elif identity.user == token_user:
        # we have a match, a user with this email and identity
        # TODO: update the identity if need be
        pass
    else:
        raise OAuthError("This identity is already claimed by another user.")

    return identity



import click
import functools
import logging
import importlib
import sys

from .utils import remove_ansi_escape_sequences


# By default, click is automatically stripping ANSI color codes if the output
#  stream is not connected to a terminal (which is the case with OutputLogger).
# In order to force it to use colors, we have to add a `color=True` param to secho
# But this turns off autodetection, so we now have to be careful not to force
#  this for Windows or other unsupported platforms. Hence the check bellow:
if sys.platform.startswith(("linux", "darwin")):
    FORCE_COLOR = True
else:
    FORCE_COLOR = None


class OutputLogger(object):
    """
    Context manager that takes over stdout and stderr and redirects anything
     sent there to a log file as also to the real stdout and stderr

    It also resets the handlers of the root logger to point to the newly set
     stderr, so that logging messages are also send to the log file
     even if it is not specified by the logger of a particular module.
    """

    def __init__(self, file):
        # Don't append, just log the last run for the same job_id
        self.file = file
        self.stdout = sys.stdout
        self.stderr = sys.stderr

        self.log_handlers = logging.getLogger().handlers

    def __enter__(self):
        # assign the Tee class to stdout and stderr so that everything pass through it
        sys.stdout = Tee(self.file, self.stdout)
        sys.stderr = Tee(self.file, self.stderr)

        stdlog = logging.StreamHandler(sys.stderr)
        logging.getLogger().handlers = [stdlog]

        # monkeypatch `click.secho`, setting `color=FORCE_COLOR` by default
        click.secho = functools.partial(click.secho, color=FORCE_COLOR)

    def __exit__(self, exc_type, exc_value, tb):
        # log any Exception that caused this OutputLogger to exit prematurely
        if isinstance(exc_value, Exception):
            logging.exception(exc_value)

        sys.stdout.flush()
        sys.stderr.flush()

        sys.stdout = self.stdout
        sys.stderr = self.stderr
        logging.getLogger().handlers = self.log_handlers

        # revert back to the `click` implementation
        importlib.reload(click)


class Tee(object):
    """
    Simple Tee class to log anything written in a stream to a file
     and then also write it to the stream.
    """

    def __init__(self, file, stream):
        self.file = file
        self.stream = stream

    def write(self, data):
        self.stream.write(data)
        self.file.write(remove_ansi_escape_sequences(data))
        # always flush the file to keep it up to date with the stream
        self.file.flush()

    def flush(self):
        self.stream.flush()
        self.file.flush()



import logging
import re
from fnmatch import fnmatch
from collections import OrderedDict, namedtuple
from enum import Enum, auto
from functools import singledispatch
from typing import List

from meltano.core.behavior.visitor import visit_with


SelectPattern = namedtuple(
    "SelectPattern", ("stream_pattern", "property_pattern", "negated")
)


def parse_select_pattern(pattern: str):
    negated = False

    if pattern.startswith("!"):
        negated = True
        pattern = pattern[1:]

    stream, *_ = pattern.split(".")

    return SelectPattern(
        stream_pattern=stream, property_pattern=pattern, negated=negated
    )


def path_property(path: str):
    """
    Extract the property name from a materialized path.

    As we traverse the catalog tree, we build a materialized path
    to keep track of the parent nodes.

    Examples:

      stream[0].properties.list_items.properties.account → list_items.account
      stream[0].properties.name                          → name
    """
    prop_regex = r"properties\.([\w\[\]\d]+)+"
    components = re.findall(prop_regex, path)
    return ".".join(components)


class CatalogNode(Enum):
    STREAM = auto()
    STREAM_METADATA = auto()
    STREAM_PROPERTY = auto()
    STREAM_PROPERTY_METADATA = auto()


class SelectionType(str, Enum):
    SELECTED = "selected"
    EXCLUDED = "excluded"
    AUTOMATIC = "automatic"

    def __bool__(self):
        return self is not self.__class__.EXCLUDED

    def __add__(self, other):
        if self is SelectionType.EXCLUDED or other is SelectionType.EXCLUDED:
            return SelectionType.EXCLUDED

        if self is SelectionType.AUTOMATIC or other is SelectionType.AUTOMATIC:
            return SelectionType.AUTOMATIC

        return SelectionType.SELECTED


@singledispatch
def visit(node, executor, path: str = ""):
    logging.debug(f"Skipping node at '{path}'")


@visit.register(dict)
def _(node: dict, executor, path=""):
    node_type = None

    if re.search(r"streams\[\d+\]$", path):
        node_type = CatalogNode.STREAM

    if re.search(r"schema(\.properties\.\w*)+$", path):
        node_type = CatalogNode.STREAM_PROPERTY

    if re.search(r"metadata\[\d+\]$", path) and "breadcrumb" in node:
        if len(node["breadcrumb"]) == 0:
            node_type = CatalogNode.STREAM_METADATA
        else:
            node_type = CatalogNode.STREAM_PROPERTY_METADATA

    if node_type:
        logging.debug(f"Visiting {node_type} at '{path}'.")
        executor(node_type, node, path)

    for child_path, child_node in node.items():
        if node_type is CatalogNode.STREAM_PROPERTY and child_path in ["anyOf", "type"]:
            continue

        # TODO mbergeron: refactor this to use a dynamic visitor per CatalogNode
        executor.visit(child_node, path=f"{path}.{child_path}")


@visit.register(list)
def _(node: list, executor, path=""):
    for index, child_node in enumerate(node):
        executor.visit(child_node, path=f"{path}[{index}]")


@visit_with(visit)
class CatalogExecutor:
    def execute(self, node_type: CatalogNode, node, path):
        dispatch = {
            CatalogNode.STREAM: self.stream_node,
            CatalogNode.STREAM_METADATA: self.stream_metadata_node,
            CatalogNode.STREAM_PROPERTY_METADATA: self.property_metadata_node,
            CatalogNode.STREAM_PROPERTY: self.property_node,
        }

        try:
            dispatch[node_type](node, path)
        except KeyError:
            logging.debug(f"Unknown node type '{node_type}'.")

    def stream_node(self, node, path: str):
        pass

    def property_node(self, node, path: str):
        pass

    def stream_metadata_node(self, node, path: str):
        pass

    def property_metadata_node(self, node, path: str):
        pass

    def __call__(self, node_type, node, path):
        return self.execute(node_type, node, path)


class SelectExecutor(CatalogExecutor):
    def __init__(self, patterns: List[str]):
        self._stream = None
        self._patterns = list(map(parse_select_pattern, patterns))

    @property
    def current_stream(self):
        return self._stream["tap_stream_id"]

    @classmethod
    def _match_patterns(cls, value, include=[], exclude=[]):
        included = any(fnmatch(value, pattern) for pattern in include)
        excluded = any(fnmatch(value, pattern) for pattern in exclude)

        return included and not excluded

    def update_node_selection(self, node, path: str, selected: bool):
        node["selected"] = selected
        if selected:
            logging.debug(f"{path} has been selected.")
        else:
            logging.debug(f"{path} has been deselected.")

    def stream_match_patterns(self, stream):
        return self._match_patterns(
            stream,
            include=(
                pattern.stream_pattern
                for pattern in self._patterns
                if not pattern.negated
            ),
            exclude=(
                pattern.stream_pattern
                for pattern in self._patterns
                if pattern.negated and pattern.property_pattern == "*"
            ),
        )

    def property_match_patterns(self, prop):
        return self._match_patterns(
            prop,
            include=(
                pattern.property_pattern
                for pattern in self._patterns
                if not pattern.negated
            ),
            exclude=(
                pattern.property_pattern
                for pattern in self._patterns
                if pattern.negated
            ),
        )

    def stream_node(self, node, path):
        self._stream = node
        selected = self.stream_match_patterns(self.current_stream)
        stream_metadata = {"breadcrumb": [], "metadata": {"inclusion": "automatic"}}

        try:
            metadata = next(
                metadata
                for metadata in node["metadata"]
                if len(metadata["breadcrumb"]) == 0
            )
            self.update_node_selection(metadata["metadata"], path, selected)
        except KeyError:
            node["metadata"] = [stream_metadata]
        except StopIteration:
            # This is to support legacy catalogs
            node["metadata"].insert(0, stream_metadata)

        # the node itself has a `selected` key
        self.update_node_selection(node, path, selected)

    def stream_metadata_node(self, node, path):
        metadata = node["metadata"]
        selected = self.stream_match_patterns(self.current_stream)
        self.update_node_selection(metadata, path, selected)

    def property_node(self, node, path):
        breadcrumb_idx = path.index("properties")
        breadcrumb = path[breadcrumb_idx:].split(".")

        try:
            next(
                metadata
                for metadata in self._stream["metadata"]
                if metadata["breadcrumb"] == breadcrumb
            )
        except StopIteration:
            # This is to support legacy catalogs
            self._stream["metadata"].append(
                {"breadcrumb": breadcrumb, "metadata": {"inclusion": "automatic"}}
            )

    def property_metadata_node(self, node, path):
        property_path = ".".join(node["breadcrumb"])
        prop = f"{self.current_stream}.{path_property(property_path)}"
        selected = self.property_match_patterns(prop)
        self.update_node_selection(node["metadata"], path, selected)


class ListExecutor(CatalogExecutor):
    def __init__(self, selected_only=False):
        # properties per stream
        self.properties = OrderedDict()

        super().__init__()

    def stream_node(self, node, path):
        stream = node["tap_stream_id"]
        if stream not in self.properties:
            self.properties[stream] = set()

    def property_node(self, node, path):
        prop = path_property(path)
        # current stream
        stream = next(reversed(self.properties))
        self.properties[stream].add(prop)


class ListSelectedExecutor(CatalogExecutor):
    SelectedNode = namedtuple("SelectedNode", ("key", "selection"))

    def __init__(self):
        self.streams = set()
        self.properties = OrderedDict()
        super().__init__()

    @property
    def selected_properties(self):
        # we don't want to mutate the visitor result
        selected = self.properties.copy()

        # remove all non-selected streams
        for stream in (name for name, selected in self.streams if not selected):
            del selected[stream]

        # remove all non-selected properties
        for stream, props in selected.items():
            selected[stream] = {name for name, selected in props if selected}

        return selected

    def node_selection(self, node):
        try:
            metadata = node["metadata"]
            if metadata.get("inclusion") == "automatic":
                return SelectionType.AUTOMATIC

            if metadata.get("selected", False):
                return SelectionType.SELECTED

            return SelectionType.EXCLUDED
        except KeyError:
            return False

    def stream_node(self, node, path):
        self._stream = node["tap_stream_id"]
        self.properties[self._stream] = set()

    def stream_metadata_node(self, node, path):
        selection = self.SelectedNode(self._stream, self.node_selection(node))
        self.streams.add(selection)

    def property_metadata_node(self, node, path):
        property_path = ".".join(node["breadcrumb"])
        prop = path_property(property_path)
        selection = self.SelectedNode(prop, self.node_selection(node))

        self.properties[self._stream].add(selection)



import json
import logging

from meltano.core.config_service import ConfigService
from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginLacksCapabilityError
from meltano.core.plugin_invoker import invoker_factory
from meltano.core.plugin.singer.catalog import ListSelectedExecutor
from .project import Project
from .db import project_engine


class SelectService:
    def __init__(
        self, project: Project, extractor: str, config_service: ConfigService = None
    ):
        self.project = project
        self.config = config_service or ConfigService(project)
        self._extractor = self.config.find_plugin(extractor, PluginType.EXTRACTORS)

    @property
    def extractor(self):
        return self._extractor

    def load_schema(self, session):
        invoker = invoker_factory(
            self.project, self.extractor, prepare_with_session=session
        )

        # ensure we already have the discovery run at least once
        if not invoker.files["catalog"].exists():
            logging.info("Catalog not found, trying to run the tap with --discover.")
            self.extractor.run_discovery(invoker)

        # update the catalog accordingly
        self.extractor.apply_select(invoker)

        # return the updated catalog
        with invoker.files["catalog"].open() as catalog:
            return json.load(catalog)

    def list_all(self, session) -> ListSelectedExecutor:
        list_all = ListSelectedExecutor()

        try:
            schema = self.load_schema(session)
            list_all.visit(schema)
        except FileNotFoundError as e:
            logging.error(
                "Cannot find catalog: make sure the tap runs correctly with --discover; `meltano invoke TAP --discover`"
            )
            raise e

        return list_all

    def select(self, entities_filter, attributes_filter, exclude=False):
        exclude = "!" if exclude else ""
        pattern = f"{exclude}{entities_filter}.{attributes_filter}"
        self.extractor.add_select_filter(pattern)
        self.config.update_plugin(self.extractor)



import json
import os
from os.path import join
from pathlib import Path

import markdown
from flask import jsonify, request

from meltano.core.project import Project
from meltano.core.compiler.project_compiler import ProjectCompiler
from meltano.api.api_blueprint import APIBlueprint
from meltano.core.m5o.m5o_file_parser import (
    MeltanoAnalysisFileParser,
    MeltanoAnalysisMissingTopicFilesError,
    MeltanoAnalysisFileParserError,
)
from meltano.core.m5o.m5o_collection_parser import (
    M5oCollectionParser,
    M5oCollectionParserTypes,
)
from flask_principal import Need
from meltano.api.security.resource_filter import ResourceFilter, NameFilterMixin
from meltano.api.security.auth import permit
from meltano.api.json import freeze_keys


reposBP = APIBlueprint("repos", __name__)


class ReportIndexFilter(NameFilterMixin, ResourceFilter):
    def __init__(self, *args):
        super().__init__(*args)

        self.needs(self.design_need)

    def design_need(self, permission_type, report):
        return Need("view:design", report["design"])


class M5ocFilter(ResourceFilter):
    def filter_designs(self, designs):
        return (
            ResourceFilter().needs(self.design_need).filter_all("view:design", designs)
        )

    def design_need(self, permission_type, design_name):
        return Need(permission_type, design_name)

    def scope(self, permission_type, m5oc):
        for topic, topic_def in m5oc.items():
            topic_def["designs"] = self.filter_designs(topic_def["designs"])

        return {
            topic: topic_def
            for topic, topic_def in m5oc.items()
            if topic_def["designs"]
        }


@reposBP.route("/", methods=["GET"])
def index():
    project = Project.find()
    onlyfiles = [f for f in project.model_dir().iterdir() if f.is_file()]

    path = project.model_dir()
    dashboardsParser = M5oCollectionParser(path, M5oCollectionParserTypes.Dashboard)
    reportsParser = M5oCollectionParser(path, M5oCollectionParserTypes.Report)

    items = reportsParser.parse()
    dashboardFiles = dashboardsParser.parse()
    reportsFiles = ReportIndexFilter().filter_all("view:reports", items)

    sortedM5oFiles = {
        "dashboards": {"label": "Dashboards", "items": dashboardFiles},
        "documents": {"label": "Documents", "items": []},
        "topics": {"label": "Topics", "items": []},
        "reports": {"label": "Reports", "items": reportsFiles},
        "tables": {"label": "Tables", "items": []},
    }
    onlydocs = project.model_dir().parent.glob("*.md")

    for d in onlydocs:
        file_dict = MeltanoAnalysisFileParser.fill_base_m5o_dict(
            d.relative_to(project.root), str(d.name)
        )
        sortedM5oFiles["documents"]["items"].append(file_dict)

    for f in onlyfiles:
        filename, ext = os.path.splitext(f)
        if ext != ".m5o":
            continue

        # filename splittext twice occurs due to current *.type.extension convention (two dots)
        filename = filename.lower()
        filename, ext = os.path.splitext(filename)
        file_dict = MeltanoAnalysisFileParser.fill_base_m5o_dict(
            f.relative_to(project.root), filename
        )
        if ext == ".topic":
            sortedM5oFiles["topics"]["items"].append(file_dict)
        if ext == ".table":
            sortedM5oFiles["tables"]["items"].append(file_dict)

    m5o_parser = MeltanoAnalysisFileParser(project)

    for package in m5o_parser.packages():
        package_files = MeltanoAnalysisFileParser.package_files(package)
        sortedM5oFiles["topics"]["items"] += package_files["topics"]["items"]
        sortedM5oFiles["tables"]["items"] += package_files["tables"]["items"]

    if not len(sortedM5oFiles["topics"]["items"]):
        return jsonify(
            {
                "result": False,
                "errors": [{"message": "Missing topic file(s)", "file_name": "*"}],
                "files": [],
            }
        )

    return jsonify(sortedM5oFiles)


def lint_all(compile):
    project = Project.find()
    compiler = ProjectCompiler(project)
    try:
        compiler.parse()
    except MeltanoAnalysisFileParserError as e:
        return jsonify(
            {
                "result": False,
                "errors": [{"message": e.message, "file_name": e.file_name}],
            }
        )
    if compile:
        compiler.compile()
    return jsonify({"result": True})


@reposBP.errorhandler(MeltanoAnalysisFileParserError)
def handle_meltano_analysis_file_parser_error(e):
    return (
        jsonify(
            {
                "result": False,
                "errors": [{"message": e.message, "file_name": e.file_name}],
            }
        ),
        500,
    )


@reposBP.errorhandler(FileNotFoundError)
def handle_file_not_found(e):
    return jsonify({"result": False, "error": str(e)}), 404


@reposBP.route("/lint", methods=["GET"])
def lint():
    return lint_all(False)


@reposBP.route("/sync", methods=["GET"])
def sync():
    return lint_all(True)


@reposBP.route("/models", methods=["GET"])
def models():
    project = Project.find()
    topicsFile = project.run_dir("models", "topics.index.m5oc")
    path = Path(topicsFile)

    topics = json.load(open(path, "r")) if path.is_file() else {}
    topics = next(M5ocFilter().filter("view:topic", [topics]))

    # for now let's freeze the keys, so the we can re-use the
    # key name as the design's name, but this should probably
    # return a list
    topics = freeze_keys(topics)
    return jsonify(topics)


@reposBP.route("/designs/<path:namespace>/<topic_name>/<design_name>", methods=["GET"])
def design_read(namespace, topic_name, design_name):
    permit("view:design", design_name)

    project = Project.find()
    topic = project.run_dir("models", namespace, f"{topic_name}.topic.m5oc")

    with topic.open() as f:
        topic = json.load(f)
    designs = topic["designs"]
    design = next(e for e in designs if e["name"] == design_name)
    return jsonify(design)



from meltano.core.project import Project
from meltano.core.schedule_service import ScheduleService

from .reports_helper import ReportsHelper
from .sql_helper import SqlHelper


class DashboardsHelper:
    def get_dashboard_reports_with_query_results(self, reports):
        project = Project.find()
        schedule_service = ScheduleService(project)
        sql_helper = SqlHelper()
        reports_helper = ReportsHelper()

        for report in reports:
            reports_helper.update_report_with_query_results(
                report, schedule_service, sql_helper
            )

        return reports
